package bitcreek;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileWriter;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.LinkedList;

/**
 * Questa classe scrive su file e stampa a monitor tutte le azioni rilevanti
 * eseguite sia dal client che dal server.<br> I metodi pi usati di questa
 * classe sono <code>out</code> e <code>err</code>, entrambi scrivono sul
 * file la stringa o l'oggetto passato come parametro, tuttavia il metodo
 * <code>out</code> stampa il messaggio sullo <code>stdout</code>, mentre
 * <code>err</code> stampa sullo <code>stderr</code>.<br>
 * La classe implementa, inoltre, un buffer di dimensione variabile
 * supportato dall'oggetto <code>LinkedList<StringBuilder> pending</code>,
 * che una volta raggiunta la sua capienza massima, viene svuotato del suo
 * contenuto effettuando la scrittura su file di tutti gli eventi.<br>
 * L'idea che sta alla base di questa scelta implementativa opzionale e' di
 * ottimizzare il numero di accessi al disco.<br><br>
 * Questa classe implementa anche due costruttori, il primo ha come
 * parametro il solo nome del file che verra' usato per memorizzare su disco
 * gli eventi del programma. Il secondo costruttore permette di definire con
 * precisione, oltre al nome del file, la cartella dove verra' memorizzato,
 * la sua dimensione massima e permette di escludere l'uso del buffer
 * imponendo la scrittura immediata di ogni evento su disco.<br>
 * Se non diversamente specificato la cartella dove il file verra'
 * memorizzato e' log, che si trovera' nella cartella radice del
 * programma.<br><br>
 * Ogni evento viene memorizzato su file comprensivo di data e orario,
 * inoltre se esiste un file gia' con lo stesso nome ogni nuovo evento
 * verra' appeso in coda al file. Se il file supera la dimensione massima
 * definita verr' creato un nuovo file (<i>nome_del_file_2.log</i>).<br> 
 * Se a sua volta questo file raggiunge la sua dimensione massima verra'
 * sovrascritto il file precedente (<i>nome_del_file_1.log</i>).
 * Nel caso che esistano entrambi i file esistano e entrambi siano saturi,
 * l'oggetto andra' a sovrascrivere sempre il file pi vecchio.<br>
 * Se si vuole forzare lo scaricamento delle informazioni contenute nel
 * buffer dell'oggetto il metodo flush serve allo scopo.<br>
 * L'override del metodo finalize obbliga l'oggetto a scrivere sul disco
 * tutte le informazioni pendenti nel buffer, benche' sia auspicabile
 * richiamare opportunamente il metodo flush prima della chiusura del
 * programma.<br>
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732<br>
 * 	     	Progetto LPR anno 2008/2009. */
public class Logger {
	private static final int			MAX_PENDING = 8*1024;
	private static final String			DEBUG_DIR_NAME = "log";
	private String 						name;
	private long						max_file_size	= 614400;	/* 600Kb */
	private File						debug_dir;
	private boolean						first_file		= true;
	private boolean						first_write		= true;
	private LinkedList<StringBuilder>	pending;
	private int							pending_size	= MAX_PENDING;
	private boolean						direct_writes	= true;
	private static boolean				loggers_enabled = true;
	
	private void createDebugDir(String debug_dir_name){
		File file = new File(debug_dir_name);
		if ((file.exists() && file.isFile())){
			file.delete();
		}
		if (!file.exists()){
			file.mkdir();
		}
		debug_dir = file;
	}
	
	/**
	 * Costruttore generico per la classe Logger
	 * @param filename Nome del file di log. */
	public Logger(String filename){
		createDebugDir(DEBUG_DIR_NAME);
		name = filename;
		
		try{
			File	f1 = getLogFile();
			first_file = false;
			File	f2 = getLogFile();			
			first_file = true;

			if (f1.exists() && f2.exists()){
				if (f1.lastModified() < f2.lastModified()){
					first_file = false;
				}
			}
		}
		catch(Throwable e){
			System.err.println(e);
		}
	}

	/**
	 * Costruttore specifico per la classe Logger
	 * @param debug_dir_name Cartella dove verranno creati i file di log.
	 * @param filename Nome del file di log.
	 * @param max_file_size Dimensione massima di ogni file di log.
	 * @param direct_writes Indica, se ipostata <code>True</code> la
	 *        scrittura immediata di ogni informazione sul disco. */
	public Logger(String debug_dir_name, String filename,
		long max_file_size, boolean direct_writes){
			name = filename;
			this.max_file_size = max_file_size;
			this.direct_writes = direct_writes;
			if (debug_dir_name==null)
				createDebugDir(DEBUG_DIR_NAME);
			else
				createDebugDir(debug_dir_name);
			
			try{
				File f1 = getLogFile();
				first_file = false;
				File f2 = getLogFile();			
				first_file = true;
			
				if (f1.exists() && f2.exists()){
					if (f1.lastModified() < f2.lastModified()){
						first_file = false;
					}
				}
			}
			catch(Throwable e){
				System.err.println(e);
			}
	}
	
	/**
	 * @param _str
	 */
	public void log(String _str){
		if (!loggers_enabled ){return;}
		
		Calendar now = GregorianCalendar.getInstance();
		StringBuilder str = new StringBuilder(_str.length() + 20);
		String timeStamp = "[" 
			+ format(now.get(Calendar.DAY_OF_MONTH))
			+ "-" + format(now.get(Calendar.MONTH)+1)
			+ " " + format(now.get(Calendar.HOUR_OF_DAY))
			+ ":" + format(now.get(Calendar.MINUTE))
			+ ":" + format(now.get(Calendar.SECOND)) + "] ";        

		synchronized(this){
			if (first_write){
				
				first_write = false;
				str.append("\r\n[");
				str.append(now.get(Calendar.YEAR));
				str.append("] Log File per BitCreek ");
				str.append(name);
				str.append("\r\n" );
				
			}
			str.append(timeStamp);
			str.append(_str);
		
			if (!direct_writes){
				if (pending == null){
					pending = new LinkedList<StringBuilder>();
				}
				pending.add(str);
				pending_size += str.length();
				if (pending_size > MAX_PENDING){
					writePending();
				}
				return;
			}
			write(str);
		}
	}
	
	private void write(StringBuilder str){
		PrintWriter	pw = null;
		
		try{	
			File log_file = getLogFile();

			if (log_file.length()>=max_file_size){
				first_file = !first_file;
				log_file = getLogFile();	
				log_file.delete();
			}
			pw = new PrintWriter(new FileWriter(log_file, true));
			pw.println(str);
		}
		catch(Throwable e){}
		finally{
			if(pw!=null){pw.close();}
		}
	}
	
	protected void writePending() {
		synchronized(this){
			if (pending == null){ return; }
			PrintWriter	pw = null;
			
			try{	
				File log_file = getLogFile();
				/**
				 *  log_file.length ritornrea' 0 se il file non esiste, cosi'
				 *  non c'e' l'esplicita necessita' di controllare la sua
				 *  esistenza. */
				if (log_file.length() >= max_file_size){
					first_file = !first_file;
					log_file = getLogFile();	
					log_file.delete();
				}
				pw = new PrintWriter(new FileWriter(log_file, true));
				for (StringBuilder str: pending){
					pw.println(str);
				}
			}
			catch(Throwable e){}
			finally{				
				direct_writes 	= true;
				pending			= null;
				if (pw!=null){pw.close();}
			}
		}
	}
	
	protected String getName(){return(name);}
	
	/** Scrive su file di log tutte le stringhe pendenti  */
	public void flush(){
		if (!direct_writes){writePending();}
	}
		
	/**
	 * @param e
	 * */
	public void log(Throwable e){
		try{
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			PrintWriter	pw = new PrintWriter(new OutputStreamWriter(baos));
			e.printStackTrace(pw);
			pw.close();
			log(baos.toString());
		}
		catch(Throwable ignore){}
	}
	
	/**
	 * @param str Le informazioni che si vogliono loggare su file e stampare
	 *            a monitor sullo standar output. */
	public void out(String str){
		System.out.println(str);
		log(str);
	}
	
	/**
	 * @param o L'oggetto di cui si vuole loggare le informazioni */
	public void out(Object o){
		System.out.println(o.toString());
		log(o.toString());
	}
	
	/**
	 * @param str Le informazioni che si vogliono loggare su file e stampare
	 *            a monitor sullo standar error. */
	public void err(String str){
		System.err.println(str);
		log(str);
	}
	
	/**
	 * @param e Throwable */
	public void err(Throwable e){
		e.printStackTrace();
		log(e);
	}
	
	private File getLogFile(){
		return(new File(debug_dir, getName() + "_"
					+ (first_file?"1":"2") + ".log"));
	}
	
	private static String format(int n){
		if (n < 10){return( "0" + n );}
	   return( String.valueOf(n));
	}
	
	@Override
	protected void finalize() throws Throwable{
		if (!direct_writes){writePending();}
		super.finalize();
	}
	
}
