package bitcreek;

import java.io.Serializable;
import java.net.InetSocketAddress;
import java.util.Arrays;
import java.util.BitSet;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Questa classe definisce in modo univoco tutti i messaggi che il client
 * scambier con gli altri client e con il server.<br>
 * Il  suo costruttore e' privato dunque per poter creare una nuova istanza
 * di questa classe bisogna usare i metodi statici definiti opportunamente
 * a seconda del tipo del messaggio che si vuole inviare.<br>
 * Ogni metodo statico definisce il tipo di messaggio attraverso l'attributo
 * code che e' univoco e attraverso questo e' possibile identificare ogni
 * messaggio che il client o il server riceve.<br> 
 * L'oggetto una volta creato e' immutabile poich ogni attributo interno
 * di esso e' definito come <code>final</code>.<br><br>
 * La scelta di mantenere una visibilit pubblica per tutti gli attributi
 * mi ha permesso di evitare di creare dei metodi '<i>get</i>' per reperire
 * le informazioni dell'oggetto, che, se da una parte aumenta lo spreco di
 * memoria (l'oggetto una volta consumato viene gettato via),
 * dall'altra aumenta sensibilmente le prestazioni del programma.
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732<br>
 * 	     	Progetto LPR anno 2008/2009. */
public class Message implements Serializable{
	private static final long serialVersionUID = -3718413986867375460L;
	/** Codice di connessione rifiutata */
	public static final byte CONN_ERR = 0;
	/** Codice di richiesta di connessione */
	public static final byte CONN_REQ = 1;
	/** Codice di connessione accettata */
	public static final byte CONN_ACK = 2;
	
	/** Codice d'errore sul blocco richiesto. (Non posseduto dal peer) */
	public static final byte BLOCK_ERR = 3;		
	/** Codice per la richiesta di un blocco */
	public static final byte BLOCK_REQ = 4;		
	/** Codice per il messaggio d'invio di un blocco */
	public static final byte BLOCK_ACK = 5;		
	/** Codice per la richiesta del BitSet dei blocchi del Peer remoto */
	public static final byte BITSET_GET = 6;
	/** Codice per l'invio del BitSet dei blocchi del Peer al remoto */
	public static final byte BITSET_SEND = 7;
	/** Codice per il KeepAlive */
	public static final byte KEEPALIVE = 8;

	/** Codice per la richieta dello swarm al Tracker TCP */
	public static final byte SWARM_REQUEST = 9;
	/** Codice di errore per Swarm trovato sul Server */
	public static final byte SWARM_FOUND = 10;
	/** Codice di errore per Swarm non trovato sul Server */
	public static final byte SWARM_NOT_FOUND = 11;
	
	/** Codice del messaggio */
	public final byte code;	
	/** Oggetto 'InetSocketAddress' che identifica il peer */
	public final InetSocketAddress idPeer;
	/** Nome del file che si richiede */
	public final String filename;
	/** Indice del blocco */
	public final int indice;
	/** Il BitSet dei blocchi che il Peer possiede */
	public final BitSet bs_blocchi;
	/** Array di byte del blocco che si sta inviando */
	public final byte[] blocco;
	/** Il costruttore della classe e' privato, la variabile swarm e' final
	 * quindi non c'e' modo di cambiarla una volta inizializzata, dunque e'
	 * possibile renderla pubblica, per un'accesso immediato
	 * all'informazione */
	public final ConcurrentHashMap<Integer, Peer> swarm;
	
	private Message(byte code, InetSocketAddress peer, String filename,
		BitSet bs_blocchi, byte[] blocco, int indice,
		ConcurrentHashMap<Integer, Peer> swarm){
			this.code = code;
			this.bs_blocchi = bs_blocchi;
			this.idPeer = peer;
			this.filename = filename;
			this.indice = indice;
			this.blocco = blocco;
			this.swarm = swarm;
	}

	/* Costruttori per l'handshake della fase di connessione */
	/** Metodo per costruire un messaggio di richiesta di connessione ad un Peer
	 * @param idPeer Oggetto 'InetSocketAddress' che identifica il peer.
	 * @param filename Il nome del file che si richiede.
	 * @param blocchi Il BitSet dei blocchi che il Peer possiede. 
	 * @return un messaggio con codice CONN_REQ, l'identificativo del peer */	
	public static Message connReq(InetSocketAddress idPeer, String filename, BitSet blocchi){
		if (idPeer==null) throw new NullPointerException();
		if (filename==null) throw new NullPointerException();
		if (blocchi==null) throw new NullPointerException();
		return new Message(CONN_REQ,idPeer,filename,blocchi,null,0,null);
	}
	
	/** Metodo per costruire un messaggio di accettata connessione con un Peer.
	 * @param idPeer Oggetto 'InetSocketAddress' che identifica il peer.
	 * @param blocchi Il BitSet dei blocchi che il Peer possiede.
	 * @return un messaggio con codice CONN_ACK, l'identificativo del peer */
	public static Message connAck(InetSocketAddress idPeer, BitSet blocchi){
		if (idPeer==null) throw new NullPointerException();
		if (blocchi==null) throw new NullPointerException();
		return new Message(CONN_ACK,idPeer,null,blocchi,null,0,null);
	}
	
	/**
	 * @param idPeer Oggetto 'InetSocketAddress' che identifica il peer.
	 * @return un messaggio con codice CONN_ERR e l'identificativo del peer. */
	public static Message connErr(InetSocketAddress idPeer){
		if (idPeer==null) throw new NullPointerException();
		return new Message(CONN_ERR,idPeer,null,null,null,0,null);
	}

	/* Costruttori per la fase di dialogo tra Peer */
	/** Metodo per costruire un messaggio di richiesta di un blocco ad un Peer.
	 * @param blockIndex L'indice del blocco richiesto.
	 * @return un messaggio con codice BLOCK_GET e l'indice del blocco richiesto */
	public static Message blockRequest(int blockIndex){
		return new Message(BLOCK_REQ,null,null,null,null,blockIndex,null);
	}
	
	/** Metodo per costruire invia il blocco richiesto al Peer
	 * @param blocco Il blocco che si sta inviando
	 * @param blockIndex L'indice del blocco che si sta inviando.
	 * @return un messaggio con codice BLOCK_SND e il blocco richiesto */
	public static Message blockSend(int blockIndex, byte[] blocco){
		if (blocco==null) throw new NullPointerException();
		return new Message(BLOCK_ACK,null,null,null,blocco,blockIndex,null);
	}
	
	/** Metodo per costruire un messaggio di errore su un blocco richiesto ad
	 * un Peer che non lo possiede. 
	 * @param blockIndex L'indice del blocco su cui si e' verificato l'errore
	 * @return un messaggio con codice BLOCK_ERR e l'indice del blocco su cui
	 * 		   si e' vericato l'errore. */
	public static Message blockErr(int blockIndex){
		return new Message(BLOCK_ERR,null,null,null,null,blockIndex,null);
	}
	
	/**
	 * Metodo per costruire un messaggio che invia il BitSet dei blocchi
	 * Peer.
	 * @param bs_blocchi_ Il BitSet dei blocchi che il Peer possiede.
	 * @return un messaggio con codice BITSET_SEND e i blocchi che il Peer
	 *         possiede. */
	public static Message sendPeerBlocks(BitSet bs_blocchi_){
		if (bs_blocchi_==null) throw new NullPointerException();
		return new Message(BITSET_SEND,null,null,bs_blocchi_,null,0,null);
	}
	
	/**  Metodo per costruire un messaggio che richiede nuovamente al Peer
	 * il BitSet dei blocchi che possiede.
	 * @return un messaggio con codice BITSET_GET.*/
	public static Message getPeerBlocks(){
		return new Message(BITSET_GET,null,null,null,null,0,null);
	}
	
	/* Costruttori per la comunicazione Client <--> Server */
	/** Metodo per costruire un messaggio per il KeepAlive 
	 * @param idPeer Oggetto 'InetSocketAddress' che identifica il peer.
	 * @param filename Il nome dello Swarm su cui si vuole effettare il KeepAlive
	 * @return un messaggio con codice KEEPALIVE, l'identificativo del 
	 * 		   peer e il nome dello swarm su cui si esegue il KeepAlive */
	public static Message keepalive(InetSocketAddress idPeer, String filename){
		return new Message(KEEPALIVE,idPeer,filename,null,null,0,null);
	}
	
	/** Metodo per costruire un messaggio per la richiesta dello Swarm al Tracker TCP 
	 * @param idPeer Oggetto 'InetSocketAddress' che identifica il peer.
	 * @param filename
	 * @return un messaggio con codice SWARM_REQUEST, l'identificativo del 
	 * 		   peer, il nome dello swarm che si richiede e i blocchi che,
	 * 		   eventualmente il Peer possiede. */
	public static Message swarmRequest(InetSocketAddress idPeer, String filename){
		return new Message(SWARM_REQUEST,idPeer,filename,null,null,0,null);
	}
	
	/** Il Tracker TCP risponde con questo messaggio per restituire lo swarm 
 	 * richiesto dal Peer.
 	 * @param swarm ConcurrentHashMap<Integer, Peer> La lista dei peer dello
 	 *        Swarm.
	 * @return un messaggio con codice SWARM_FOUND, e la lista dei peer
	 *        dello swarm richiesta dal Peer.*/
	public static Message swarmFound(ConcurrentHashMap<Integer, Peer> swarm){
		return new Message(SWARM_FOUND,null,null,null,null,0,swarm);
	}

	/** Il Tracker TCP risponde con questo messaggio quando non trova lo swarm 
 	 * richiesto dal Peer.
	 * @return un messaggio con codice SWARM_NOT_FOUND, per segnalare al peer
	 * 		   che lo sworm richiesto e' inisistente. */
	public static Message swarmNotFound(){
		return new Message(SWARM_NOT_FOUND,null,null,null,null,0,null);
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "Message hashcode = "+this.hashCode()+" [code= " + code +
				", bitset_blocchi= "+ bs_blocchi + ", blocco="
				+ Arrays.toString(blocco) +  ", filename="
				+ filename + ", idPeer=" + idPeer + ", indice=" + indice
				+ ", swarm=" + swarm + "]";
	}
}
