package bitcreek;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.Serializable;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;

/**
 * Questa classe definisce e crea oggetti che descrivono i file pubblicati
 * sul Server.<br> In particolare implementa un array dei digest del file il
 * cui numero e' calcolato in base alla dimensione del file stesso diviso il
 * dimensione di default di un blocco (tutti gli arrotondamenti vengono
 * fatti per eccesso).<br>
 * Il blocco e' l'unita' minima di scaricamento dei pezzi del file, la sua
 * dimensione, come da specifica e' di 4kb; per ogni blocco viene calcolato
 * il checksum tramite l'algoritmo SHA1.<br>
 * Per ogni blocco scaricato verra' calcolato il checksum e confrontato con
 * quello memorizzato, solo se i due valori coincidono il blocco viene
 * ritenuto corretto. La funzione che esegue questo controllo e' la
 * <code>checkBlock</code>.<br>
 * La classe mantiene anche il nome del file e la sua dimensione.<br>
 * Per scelta progettuale non e' possibile pubblicare sul Server due file,
 * anche diversi, che hanno lo stesso nome.<br>
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732<br>
 * 	     	Progetto LPR anno 2008/2009. */
public class Torrent implements Serializable{
	private static final long serialVersionUID = 8341194452832033491L;
	/** Il percorso assoluto del file */
	public final String percorso;
	/** Nome del file */
	public final String name;
	/** Lunghezza del file */
	public final long length;
	/** Numero dei blocchi del file */
	private int blocks = 0;
	/**
	 * Array dei digest del file, ogni digest ha una dimensione di 20 byte
	 * */
	private byte[][] digest = null;				
	/** Dimesione standard di ogni blocco del file (4K) */
	public static final int blockSize = 4096;
	/**
	 * La porta del Tracker TCP restituita dal Server, dove quest'ultimo e'
	 * in ascolto per soddisfare le richieste di invio dello swarm del file.
	 * */
	private int portaUDP = 0;
	/**
	 * La porta del Tracker UDP restituita dal Server, dove quest'ultimo e'
	 * in ascolto per il KeepAlive. */
	private int portaTCP = 0;
	
	/**
	 * @param file
	 * @throws NoSuchAlgorithmException
	 * @throws IOException */
	public Torrent(File file)
	throws NoSuchAlgorithmException, IOException{
		MessageDigest md = MessageDigest.getInstance("SHA-1");
		long offset = 0;
	    int numRead = 0;
	    
	    /* Apertura del file in modalita' binaria */
		FileInputStream fis = new FileInputStream(file);
		/* Salvo il nome del file */
		percorso = file.getAbsolutePath();
		name = file.getName();
		/* Salvo la dimensione del file */
		length = file.length();
		/* Calcolo il numero dei blocchi */
		blocks =(int)(length/blockSize);
	    if ((length%blockSize)!=0) blocks++;
	    /* Ogni digest ha una dimensione standard di 20 byte */
	    digest = new byte[blocks][20];
			    
	    for (int i=0; offset < length; i++){	
	    	byte[] blocco = new byte[blockSize];
	       	numRead=fis.read(blocco, 0, blockSize);
	       	md.update(blocco);
		    digest[i] = md.digest();
		   	offset += numRead;
	    }
	    fis.close();
	}
			
	/** Setta la porta UDP e TCP a cui il torrent e' stato assegnato
	 * @param portaTCP
	 * @param portaUDP */
	public synchronized void setTrackerPort(int portaTCP, int portaUDP){
		this.portaTCP = portaTCP;
		this.portaUDP = portaUDP;
	}
	
	/** 
	 * Controlla la corrtettezza del blocco scaricato.
	 * @param blocco
	 * @param index Indice del blocco
	 * @return <code>True</code> se l'SH1 calcolato sul blocco 'index'
	 * 		e' uguale a quello del descrittore del file. 
	 * @throws NoSuchAlgorithmException */
	public boolean checkBlock(int index, byte[] blocco)
	throws NoSuchAlgorithmException{
		MessageDigest md = MessageDigest.getInstance("SHA-1");
		md.update(blocco);
		byte[] digest = md.digest();
		return (Arrays.equals(this.digest[index], digest));
	}
	
	/** Ritorna il numero di porta del tracker TCP
	 * @return Il numero di porta del tracker TCP */
	public int getTrackerTCPPort(){return portaTCP;}
		
	/** Ritorna il numero di porta del tracker UDP
	 * @return Il numero di porta del tracker UDP */
	public int getTrackerUDPPort(){return portaUDP;}
	
	/** Ritorna il numero di blocchi del torrent
	 * @return Il numero di blocchi del torrent */
	public int numBlocchi(){return blocks;}
}

