package bitcreek.client;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.rmi.ConnectException;
import java.rmi.NoSuchObjectException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.StringTokenizer;
import sun.misc.Signal;
import sun.misc.SignalHandler;

import bitcreek.Logger;
import bitcreek.server.ServerRMIInterface;

/**
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732<br>
 * 	     	Progetto LPR anno 2008/2009. */
public class Client implements SignalHandler {
	private static final String err1 = "[CLIENT ERROR] Nessun parametro" +
		" passato al programma!\n               Il primo parametro deve" +
		" essere un file di testo con le operazioni che il client deve" +
		" eseguire.";
	/** Cartella che conterra' i file di log */
	private static final String LOG_DIR 		= "log";
	/** Dimensione massima di ogni file di log */
	private static final long MAX_SIZE			= 409600;	/* 400Kb */
	/** Parola iniziale di ogni file di log */
	private static final String LOG_FILENAME	= "Client";	
	/** Nome dello Stub */
	private static final String STUB_NAME 		= "BitCreek";
	/** Oggetto per loggare su file e stampare a monitor ogni evento
	 *  o azione, e il loro relativo esito, intrapresa dal client. */
	public static Logger log;
	/** Arresta il client, se settata a true */
	protected volatile static boolean shutdown = false;
	
	/**
	 * Tempo di attesa di default tra un KeepAlive e un altro espresso
	 * in millisecondi. */
	protected static final int KEEPALIVE_DEF = 10000;	/* 10 sec. */
	/** Tempo di attesa tra un KeepAlive e un altro espresso in secondi.
	 * Questa varibile vien correttamente settata nel main del programma */
	protected static int keepAliveTime;
	/** Numero iniziale di thread */
	protected static final int THREAD = 10;
	/** Numero massimo di thread */
	protected static final int MAX_THREAD = 20;
	/** Timeout sulle socket */
	protected static final int MAX_TIMEOUT = 60000;
	/** Numero massimo di connessioni in ingresso e uscita per ogni swarm */
	/* Preferibilemte settare questa varibile con valori pari. */
	protected static final int MAX_CONN = 10;
	/** Numero massimo di connessioni in ingresso */
	protected static int MAX_CONN_IN = MAX_CONN/2;
	/** Numero massimo di connessioni in uscita */
	protected static int MAX_CONN_OUT = MAX_CONN/2;
	/** Se settato a true impone la scrittura diretta su disco di ogni
	 *  blocco scaricato */
	protected static boolean DIRECT_WRITES = false;
	private static ClientImpl client = null;
    private SignalHandler oldHandler;

    /* Gestione dei segnali */
    /** 
     * Installa il gestore dei segnali per il segnale <code>signalName</code>
     * passato alla funzione.
     * @param signalName Codice del segnale che si vuole gestire.
     * @return SignalHandler */
    public static SignalHandler install(String signalName) {
        Signal diagSignal = new Signal(signalName);
        Client instance = new Client();
        instance.oldHandler = Signal.handle(diagSignal, instance);
        return instance;
    }

    public void handle(Signal signal) {
    	log.err("[CLIENT] Segnale "+signal+" catturato.");
        try {
            signalAction(signal);
            if (oldHandler != SIG_DFL && oldHandler != SIG_IGN) {
                oldHandler.handle(signal);
            }
        }
        catch (Exception e) {
        	log.err("[CLIENT ERROR] Signal handler failed, reason "
        			+ e.getMessage());
            log.err(e);
        }
    }

    /**
     * Gestore dei segnali definito dall'utente per i segnali che si
     * vogliono gestire in modo specifico. 
     * @param signal Il segnale ricevuto. */
    public void signalAction(Signal signal) {
		try {
			shutdown = true;
			if (client!=null){
				client.shutdown();
				UnicastRemoteObject.unexportObject(client, false);
			}
		}
		catch (NoSuchObjectException ignore){ ; }
        log.out("[CLIENT] Programma terminato (signalAction).");
		log.flush();
    }
    
	/**
	 * @param args */
	public static void main(String[] args){
		ServerRMIInterface server = null;
		InetAddress serverAddr = null;
		Registry remoteRegistry = null;
		int portaRMI = 0;
		
		try{
			log = new Logger(LOG_DIR,LOG_FILENAME,MAX_SIZE,true);
			/** Questi segnali sono gestiti correttamente
			 *  sia su piattaforma windows che linux. */
			Client.install("TERM");
			Client.install("INT");
			
			if (args.length>3) throw
				new IllegalArgumentException(
						"Troppi parametri passati da riga di comando!");
			
			if (args.length==3){
				portaRMI = Integer.parseInt(args[2]);
				serverAddr = InetAddress.getByName(args[1]);
				if (portaRMI < 1024 || portaRMI > 65535)
					throw new IllegalArgumentException(
							"Valore non compreso nel range 1025-65535.");	
			}
			else
				portaRMI = java.rmi.registry.Registry.REGISTRY_PORT;
			
			if (args.length==2)
				serverAddr = InetAddress.getByName(args[1]);
			else
				if (args.length==0)
					throw new ArrayIndexOutOfBoundsException();
			
			File file = new File(args[0]);
			log.flush();
			/* Un file di log specifico a seconda del  
			 * file delle azioni pasasto al client */
			String filename = LOG_FILENAME+"_"+file.getName();
			log = new Logger(LOG_DIR,filename,MAX_SIZE,true);
			
			/* Recupero e setto il KeepAlive */
			String keepAlive = System.getProperty("keepalive");
			if (keepAlive!=null){
				try{ keepAliveTime = Integer.parseInt(keepAlive); }
				catch(NumberFormatException e){
					keepAliveTime = KEEPALIVE_DEF;
				}
			}
			if (keepAliveTime<=0)
				keepAliveTime = KEEPALIVE_DEF;			
						
			/* Recupero e setto il DIRECT_WRITES */
			String directwrites = System.getProperty("directwrites"); 
			if (directwrites!=null)
				DIRECT_WRITES = Boolean.getBoolean(directwrites);
			
			if (serverAddr==null) 
				throw new IllegalArgumentException(
						"Specificare l'indirizzo del Server RMI.");
			
			if (!file.isFile()) throw
				new IllegalArgumentException("Il parametro passato da riga"+
						" di comando non esiste o non e' un file.");

			/* Recupera il registro remoto */
			remoteRegistry =
				LocateRegistry.getRegistry(serverAddr.getHostName(),portaRMI);

			/* Recupera lo stub del Server */
			server = (ServerRMIInterface) remoteRegistry.lookup(STUB_NAME);
			/* Avvia il peer */
			client = new ClientImpl(server, serverAddr);
			/* Parsing del file dei comandi */
			parsing(args[0]);
			
			log.out("[CLIENT] Arresto in corso...");
		}
		catch (ArrayIndexOutOfBoundsException e){
			log.err(err1);
		}
		catch (IllegalArgumentException e){
			log.err("[CLIENT ERROR] " +e.getLocalizedMessage());
			log.err(e);
		}
		catch (UnknownHostException e){
			log.err("[CLIENT ERROR] Host non trovato! "+e.getMessage());
		}
		catch (ConnectException e){
			log.err("[CLIENT ERROR] "+e.getMessage());
		}
		catch (NotBoundException e){
			log.err("[CLIENT ERROR] Stub non trovato: "+e.getMessage());
			log.err(e);
		}
		catch (Exception e){
			log.err("[CLIENT ERROR] "+e.getMessage());
			log.err(e);
		}
		finally{
			if (client != null && !shutdown)
				try {
			    	shutdown = true;
					client.shutdown();
					UnicastRemoteObject.unexportObject(client, false);
				}
				catch (RemoteException e){log.err(e);}
			log.out("[CLIENT] Programma terminato (main).");
			log.flush();
		}
		return;
    }
	
	/**
	 * Funzione statica che esegue il parsing del file dei comandi che il 
	 * client eseguira'.
	 * @param arg Nome del file dei comandi 
	 * @throws Exception */
	private static void parsing(String arg)
	throws Exception{
		String comando = null;
		String word = null;
		StringTokenizer token = null;
		BufferedReader in = new BufferedReader(new FileReader(arg));

		for (String line = in.readLine(); line!=null && !shutdown; line=in.readLine()){
			/* Salta i commenti */
			if (line.startsWith("#")) continue;
			try{
		       token = new StringTokenizer(line);
		       /* Salto le righe vuote */
		       if (!token.hasMoreTokens()) continue;
		       comando = token.nextToken();
		       if(comando.equalsIgnoreCase("aspetta")){
		    	   if (token.countTokens()==0) throw new Exception(
		    	   		"Il comando aspetta prevede come paramatro un numero"
		    		   	+" (long) che esprime,\nin millesimi di secondo, il"
		    		   	+" tempo che il programma aspettera' prima di"
		    		   	+" eseguire il prossimo comando.");
		    	   long time = Long.parseLong(token.nextToken()); 
		    	   log.out("[CLIENT] Aspetto \'"+time
		    			+"\' millis sec per la prossima istruzione.");
		    	   Thread.sleep(time);
		       }
		       else if (comando.equalsIgnoreCase("pubblica")){
		    	   if (token.countTokens()==0) throw new Exception(
		    			"Il comando pubblica prevede come paramatro il nome"
		    			+" di un file.");
		   			word = token.nextToken();
		   			while (token.hasMoreTokens())
		   				word += " " + token.nextToken();
		    	   client.pubblica(word);
		       }   
		       else if (comando.equalsIgnoreCase("recupera")){
		    	   if (token.countTokens()==0) throw new Exception(
		    		   "Il comando recupera prevede come paramatro il nome"
		    		   +" di un file.");
		   			word = token.nextToken();
		   			while (token.hasMoreTokens())
		   				word += " " + token.nextToken();
		    	   client.retrieve(word);
		       }
		       else if (comando.equalsIgnoreCase("rimuovi")){
		    	   if (token.countTokens()==0) throw new Exception(
		    		   "Il comando rimuovi prevede come paramatro il nome"
		    		   +" di un file.");
		   			word = token.nextToken();
		   			while (token.hasMoreTokens())
		   				word += " " + token.nextToken();   
		    	   client.removeDescriptor(word);
		       }
		       else
		    	   throw new Exception(
		    			   "Comando '"+comando+"' non riconosciuto");
			}
			catch(Exception e){
				log.err("[CLIENT ERROR] "+e.getLocalizedMessage());
				log.err(e);
			}
		}
		in.close();
	}
}
