package bitcreek.client;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.BindException;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;
import java.net.ServerSocket;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import bitcreek.Torrent;
import bitcreek.client.Listener.MessageIn;
import bitcreek.server.ServerRMIInterface;

/**
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732<br>
 * 	     	Progetto LPR anno 2008/2009. */
public class ClientImpl extends UnicastRemoteObject implements ClientInterface {
	private static final long serialVersionUID = -2649553668091546023L;
	/** Pool di thread per gli Swarms e il Listener  */
	protected static ThreadPoolExecutor thWorker;
	/** HashTable dei descrittori di file */
	protected final Hashtable<String, Torrent> torrentList
		= new Hashtable<String, Torrent>();
	/** Lista degli Swarms del client */
	protected static Hashtable<String, ClientSwarm> swarmsList
		= new Hashtable<String, ClientSwarm>();
	/** Lista ad accesso sincronizzato delle callback */
	private List<String> callback =
		Collections.synchronizedList(new Vector<String>());
	/** Indirizzo del Server */
	protected InetAddress serverAddress;
	/** Stub dell'oggetto remoto */
	private ServerRMIInterface serverRMI;
	protected static ServerSocket socketTCP;
	/** Coppia Indirizzo e porta che identifica univocamente il Peer */
	protected static InetSocketAddress myId;
	private Listener transfer;
	
	/** Costruttore della classe.
	 * @param server Interfaccia RMI del Server
	 * @param serverAddress Indirizzo InetAddress del Server
	 * @throws IOException */
	protected ClientImpl(ServerRMIInterface server, InetAddress serverAddress)
	throws IOException{
		if (serverAddress==null) throw new NullPointerException(
				"[CLIENT ERROR] L'indirizzo del server non e' corretto.");
		this.serverAddress = serverAddress;
		this.serverRMI = server;
		
		InetAddress[] myIpList = getComputerAddress();
		socketTCP = getSocket();
		myId = new InetSocketAddress(myIpList[0], socketTCP.getLocalPort());
		thWorker = new ThreadPoolExecutor(Client.THREAD, Client.MAX_THREAD,
				25000L, TimeUnit.MILLISECONDS,
				new LinkedBlockingQueue<Runnable>());
		thWorker.prestartCoreThread();
		transfer = new Listener();
		Client.log.out("[CLIENT] Avviato su "+myIpList[0]+":"
				+socketTCP.getLocalPort()+" [KeepAlive = "
				+Client.keepAliveTime+"]");
		thWorker.execute(new KeepAlive(serverAddress));	
		thWorker.execute(transfer);
	}
	
	/**
	 * Restituisce un array con tutti gli indirizzi IpV4 delle schede di
	 * rete del computer locale, raggiungibili e che non siano indirizzi
	 * loopback.
	 * <br>Risolve il problema per Ubuntu che dalla distro 8.04 la chiamata 
	 * <code>InetAddress.getLocalHost()</code> restituisce l'indirizzo di
	 * loopback, per un errata configurazione del file /etc/host.
	 * @see java.net.InetAddress#getLocalHost()
	 * @return Un array di InetAddress con tutti gli indirizzi IpV4 delle
	 * schede di rete del computer locale, raggiungibili e non che siano
	 * indirizzi loopback. 
	 * @throws IOException 
	 * @throws UnknownHostException */
	public static InetAddress[] getComputerAddress() throws IOException,
	UnknownHostException{
		ArrayList<InetAddress> myIpList = new ArrayList<InetAddress>();
		
		Enumeration<NetworkInterface> nets =
			NetworkInterface.getNetworkInterfaces();
		for (NetworkInterface netint : Collections.list(nets)){
			Enumeration<InetAddress> inetAddresses =
				netint.getInetAddresses();
			for (InetAddress inetAddr : Collections.list(inetAddresses)){
				if (!(inetAddr instanceof Inet6Address) && 
						(!inetAddr.isLoopbackAddress()) && 
						inetAddr.isReachable(10000)) {
					myIpList.add(inetAddr);
				}
			}
		}
		if (myIpList.isEmpty())
			throw new UnknownHostException("[CLIENT ERROR] Impossibile" +
					" determinare l'indirizzo del client!");
		return myIpList.toArray(new InetAddress[myIpList.size()]);
	}
	
	/** Restituisce un ServerSocket valido 
	 * @return Un ServerSocket valido. 
	 * @throws IOException */
	private static ServerSocket getSocket() throws IOException{
		for (int i = 2025; i<65535; i++){
			try { return new ServerSocket(i); }
			catch (BindException ignore){} 
			catch (SocketException ignore){}
		}
		throw new BindException("Nessuna porta disponibile (!?)"); 
	}
	
	/** Termina dolcemente l'esecuzione del client 
	 * <li>Termina il Thread per il KeepAlive UDP</li>
	 * <li>Termina il Thread per il trasferimento dei file tra peer</li> 
	 * <li>Elimina tutti i callback registrati sul server.</li>*/
	public void shutdown(){
		Client.log.out("[CLIENT] Arresto in corso...");
		Client.shutdown = true;
		transfer.shutdown();
		thWorker.shutdown();
		try {
			if (!thWorker.awaitTermination(15, TimeUnit.SECONDS)){
				thWorker.shutdownNow(); /* Cancel currently executing tasks */
				/* Wait a while for tasks to respond to being cancelled */
				if (!thWorker.awaitTermination(15, TimeUnit.SECONDS))
					Client.log.err("[CLIENT] Il Pool di thread non e'"+
						" terminato correttamente.");
			}
			else
				Client.log.out("[CLIENT] Pool di thread terminato.");
		}
		catch (InterruptedException ie){
			/* (Re-)Cancel if current thread also interrupted */
			thWorker.shutdownNow();
			/* Preserva l'interrupt status */
			Thread.currentThread().interrupt();
		}
		String name = null;
		while (!callback.isEmpty()){
			name = callback.remove(0);
			try { serverRMI.unregisterForCallback(name, this); }
			catch (RemoteException e){
				Client.log.err("[CLIENT] Impossibile eseguire" +
					" 'unregisterForCallback' per il file '"+name
					+"'.\n Si ha la certezza che il server sia attivo e"
					+" raggiungibile?");
				/*Client.log.err(e);*/
			}
		}
		
		/* Chiudo, per ogni swarm, tutte le comunicazioni stabilite dai
		 * Peer remoti verso questo client */
		Enumeration <ClientSwarm> swarms = swarmsList.elements();
		for (ClientSwarm swarm : Collections.list(swarms)){
			Enumeration <MessageIn> skPeer = swarm.skPeers.elements();
			for (MessageIn talk: Collections.list(skPeer)){
				talk.shutdown();
			}
		}
	}
	
	/**
	 * Rimuove un descrittore in locale e il suo swarm.
	 * Se il client aveva pubblicato il file rimuove la callback ad esso
	 * associata
	 * @param filename Nome del descrittore da rimuovere
	 * @return Ritorna true se rimuove correttamente il torrent. 
	 * @throws RemoteException */
	public boolean removeDescriptor(String filename)
	throws RemoteException{
		if (torrentList.containsKey(filename)){
			/* Elimina il descrittore */
			torrentList.remove(filename);
			
			/* Cancella lo swarm associato al descrittore */
			if (swarmsList.containsKey(filename)){
				ClientSwarm swarm = swarmsList.get(filename);
				swarmsList.remove(filename);
				/* Arresto lo swarm se questo e' attivo ovvero
				 * se sto scaricando il file. */
				if (swarm!=null) swarm.shutdown = true;
			}
			
			/* Cancella la registrazione per il callback */
			if (callback.contains(filename)){
				callback.remove(filename);
				serverRMI.unregisterForCallback(filename, this);				
			}
			Client.log.out("[CLIENT] Descrittore '"+filename
					+"' rimosso con successo!");
			return true;
		}
		else
			Client.log.err("[CLIENT] ERRORE Descrittore '"+filename
					+"' non trovato! (unregisterForCallback)");
		return false;
	}
	
	/** Crea e pubblica un nuovo descrittore di file sul server
	 * @param filename Il percorso e nome del descrittore di file da creare e
	 *        pubblicare.
	 * @return Ritorna True se la pubblicazione del file e' andata a buon fine.
	 * @throws IOException
	 * @throws NoSuchAlgorithmException Se l'algoritmo SH1 non e' presente
	 *         in questa implementazione di Java */
	public boolean pubblica(String filename) throws IOException, NoSuchAlgorithmException{
		Torrent tor = null;
		Torrent tor_tmp = null;
		
		try{ tor = new Torrent(new File(filename)); }
		catch(FileNotFoundException e){
			Client.log.err("[CLIENT ERROR] "+e.getMessage());
			return false;
		}
		catch (NoSuchAlgorithmException e){
			Client.log.err("[CLIENT ERROR] Algoritmo di codifica non "
					+"trovato! "+e.getMessage());
			return false;
		}
		
		/* Controllo che non ho un torrent gia' con quel nome in locale */
		if (!torrentList.containsKey(tor.name)){
			/* Pubblico il nuovo torrent sul server */
			tor_tmp = serverRMI.publishDescriptor(myId, tor); 
			if (tor_tmp != null){
				tor = tor_tmp;
				/* Aggiungo il torrent nella lista locale dei torrent */
				torrentList.put(tor.name, tor);
				FileCache filecache =
					FileCache.createFileCacheFromLocalTorrent(tor,1000,Client.DIRECT_WRITES);
				swarmsList.put(tor.name,
						new ClientSwarm(serverAddress,filecache));
				serverRMI.registerForCallback(tor.name, this);
				callback.add(tor.name);
				Client.log.out("[CLIENT] CallBack per il file '"
						+tor.name+"' registrata.");
				Client.log.out("[CLIENT] Descrittore '"+tor.name+
						"' pubblicato sul server.");
				return true;
			}
			else
				Client.log.err("[CLIENT] ERRORE Pubblicazione" +
					" descrittore '"+tor.name+"' sul server.");
		}
		else
			Client.log.err("[CLIENT] ERRORE Descrittore '"
				+tor.name+"' gia' presente! (pubblicazione)"); 
		return false;
	}
	
	public void notifyMe(String filename, InetSocketAddress idPeer)
	throws RemoteException{
		Client.log.out("[CLIENT CALLBACK] Il Peer "+idPeer+
				" ha scaricato il descrittore del file '"+filename+"'.");

		if (!torrentList.containsKey(filename))
			Client.log.err("[CLIENT CALLBACK ERROR] Descrittore '"
				+filename+"' non presente (Peer "+idPeer+").");
	}
	
	public InetSocketAddress getId() throws RemoteException{
		return myId;
	}
	
	
	/** Recupera un descrittore di file dal Server.
	 * @param filename Il nome del file che si richiede al Server.
	 * @return True se il file richiesto al server e' stato ricevuto,
	 *         false in tutti gli altri casi.
	 * @throws IOException */
	public boolean retrieve(String filename) throws IOException{
		Torrent tor = null;
		ClientSwarm swarm = null;
		
		/* Controllo che il descrittore del file non sia gia' presente */
		if (torrentList.containsKey(filename)){
			Client.log.err("[CLIENT] ERRORE Descrittore '"+filename
				+"' gia' presente! (retrive)");
			return false;
		}
		
		/* Prelevo il descrittore di file dal server */
		if ((tor = serverRMI.getDescriptor(filename, myId))!=null){
			Client.log.out("[CLIENT] Descrittore '"+filename
				+"' ricevuto con successo!");
			try {
				/* Se il file e' stato gia' scaricato, anche solo in parte,
				 * l'ogetto filecache se ne accorge e setta i blocchi che
				 * sono gia' si possiedono. */
				
				/* Creo un nuovo swarm.*/
				swarm = new ClientSwarm(serverAddress,
						FileCache.createFileCacheFromRemoteTorrent(tor,1000,Client.DIRECT_WRITES));
				
				/* Recupero i Peer dello Swarm */
				if (swarm.getSwarmFromServer()){
					torrentList.put(filename, tor);
					swarmsList.put(tor.name, swarm);
					/* Lancio il Thread per lo scaricamento dei
					 * blocchi del file  (se non sono gia' un seeder !!) */
					if (!swarm.filecache.testSeeder())
						thWorker.execute(swarm);
				}
			}
			catch (NoSuchAlgorithmException e){
				Client.log.err("[CLIENT] ERRORE Algoritmo di codifica" +
					" non trovato!");				
				Client.log.err(e);
				return false;
			}
			catch (IOException e){
				Client.log.err("[CLIENT] ERRORE Lettura file dal disco.");
				Client.log.err(e);
				return false;
			}
		}
		else
			Client.log.err("[CLIENT] ERRORE Descrittore '"
				+filename+"' non trovato! (retrive)");
		return false;
	}
}
