package bitcreek.client;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.security.KeyStore;
import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManagerFactory;

import bitcreek.Message;
import bitcreek.Peer;
import bitcreek.client.Listener.MessageIn;

/**
 * Questa classe mantiene tutte le informazioni riguardanti lo swarm di un
 * file. In particolare scarica dal Server lo swarm del file, e se non si e'
 * gia' un seeder, crea una lista dei Peer e dei blocchi che si devono
 * scaricare per completare il file.<br>
 * Ogni Peer dello swarm viene contattato inizialmente per la richiedergli
 * il bitset dei blocchi che possiede.<br> 
 * Eseguita questa procedura iniziale di <i>WarmUp</i>, i Peer dello swarm
 * vengono ordinati per numero di blocchi posseduti e contattati nell'ordine
 * fino che non si e' saturato il numero massimo di connessioni in uscita.
 * <br>
 * Se il numero dei Peer dello swarm e' inferiore al numero massimo di
 * connessioni in uscita, verranno stabilite tante connessioni quanti sono i
 * Peer presenti nello swarm.<br>
 * Se per un qualsiasi motivo un Peer remoto esce dallo swarm di questo 
 * client, vuoi perche' di sua iniziativa ha chiuso la connessione, vuoi
 * perche' e' stato rimosso dallo swarm per iniziativa del client stesso, 
 * se, in seguito a questo evento, si e' liberata una connessione in uscita
 * lo swarm controllera' l'esistenza di un Peer con cui e' possibile 
 * stabilire una connessione e lo contattera'.<br>
 * Se lo swarm locale di questo client si svuota, quest'ultimo contattera'
 * il Server per richiedergli nuovamente lo swarm associato a questo file,
 * sempre se ne esiste ancora uno.<br>
 * Come si e' accennato prima, questa classe crea una lista dei blocchi da
 * scaricare, che vengono ordinati per occorrenza di ogni blocco.<br>
 * In particolare, un blocco posseduto da un solo Peer verra' scaricato prima
 * rispetto ad un blocco posseduto da due o piu' Peer.<br>
 * Quando un blocco viene richiesto ad un Peer questo viene settato il flag
 * <i>download</i> a true al fine che non sia richiesto a piu' Peer
 * contemporaneamente.<br> Se la ricezione del blocco e' andata a buon fine
 * il blocco viene rimosso dalla lista dei blocchi da scaricare, altrimenti
 * il flag <i>download</i> viene reimpostato a false e il blocco verra'
 * richiesto ad un altro Peer che lo possiede, se questo esiste.<br><br>
 * La comunicazione con un Peer remoto per la richiesta e lo scaricamento dei
 * blocchi viene instaurata dalla classe <code>PeerTalk</code> il cui thread
 * viene lanciato da questa classe.  
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732<br>
 * 	     	Progetto LPR anno 2008/2009. */
public class ClientSwarm implements Runnable{
	/**
	 * @author  Giulio Tranchida (gtranchi) no. matricola 241732
	 * 	     	Progetto LPR anno 2008/2009. */
	protected class PriorityList implements Comparable<Object>{
		/** Il numero dei Peer che possiedono il blocco <code>block</code>.*/
		private int count = 0;
		/** L'indice del blocco (informazione rindondante immutabile) */
		protected final int indexBlock;
		
		/**
		 * Se impostato a true il blocco e' stato richiesto ad un Peer
		 * remoto */
		protected AtomicBoolean download = new AtomicBoolean();
		
		/** Costruttore
		 * @param indexBlock L'indice del blocco */
		protected PriorityList(int indexBlock){
			this.indexBlock = indexBlock;
		}
		
		public int compareTo(Object obj) {
			PriorityList peers = (PriorityList) obj;
			int count = peers.count;
			return (this.count<count) ? -1 : (this.count>count) ? 1 : 0;
		}
		
		/**
		 * @see java.lang.Object#hashCode() */
		@Override
		public int hashCode(){
			final int prime = 31;
			int result = 1;
			result = prime * result + getOuterType().hashCode();
			result = prime * result + indexBlock;
			return result;
		}

		/**
		 * @see java.lang.Object#equals(java.lang.Object) */
		@Override
		public boolean equals(Object obj) {
			if (this == obj) return true;
			if (obj == null) return false;
			if (!(obj instanceof PriorityList)) return false;
			PriorityList other = (PriorityList) obj;
			if (!getOuterType().equals(other.getOuterType())) return false;
			if (indexBlock != other.indexBlock) return false;
			return true;
		}

		private ClientSwarm getOuterType() {
			return ClientSwarm.this;
		}
	}
	
	/*                            CLIENT SWARM							*/
	/**
	 * Struttura dati che mantiene l'ordinamento in base alla posizione
	 * dei blocchi del file */
	private ConcurrentHashMap<Integer, PriorityList> blocksList = null;
	
	/** Struttura dati che mantiene l'ordinamento in base alla rarita'
	 * dei blocchi del file */
	private Vector<PriorityList> priorityList = null;
	
	/** Lista dei Peer dello Swarm */
	private ConcurrentHashMap<Integer, PeerTalk> swarm = null;
	
	/** 
	 * Lista delle 'socket' dei Peer che hanno richiesto una connessione
	 * remota con questo Client, per chiedere blocchi del file di questo
	 * Swarm. */
	protected ConcurrentHashMap<Integer, MessageIn> skPeers = 
		new ConcurrentHashMap<Integer, MessageIn>();

	/** Numero dei blocchi del torrent */
	protected final int nBlocchi;
	
	/** Indirizzo del Server */
	protected final InetAddress server;
	
	/** Classe che gestisce il file che si sta scaricando */
	protected final FileCache filecache;
	
	/**
	 * Arresta il thread quando quest'ultimo ha terminato lo
	 * scaricamento del file */
	protected boolean shutdown = false;
	
	/** Pool di thread per le comunicazioni in ingresso e uscita  */
	protected static ThreadPoolExecutor thCommunications = null;
	
	/** Mantiene il conteggio del numero di connessioni aperte in uscita */
	protected AtomicInteger connection;
	
	/** Costruttore della classe. 
	 * @param server L'indirizzo del Server.
	 * @param filecache Oggetto che gestisce la scrittura sul disco del
	 *        file che si sta scaricando. */
	public ClientSwarm(InetAddress server, FileCache filecache){
		this.server = server;
		this.filecache = filecache;
		this.nBlocchi = filecache.getNumBlocchi();
		
		if (!filecache.testSeeder()){
			thCommunications = new ThreadPoolExecutor(Client.MAX_CONN_OUT, 
					Client.MAX_CONN_OUT+1, 50000L, TimeUnit.MILLISECONDS,
					new LinkedBlockingQueue<Runnable>());
			swarm = new ConcurrentHashMap<Integer, PeerTalk>();
			/* Conteggio i blocchi del file che mi mancano */
			BitSet blk = (BitSet) filecache.getMyBlocks().clone();
			blk.flip(0, nBlocchi);
			int dim = blk.cardinality();
			blocksList = new ConcurrentHashMap<Integer, PriorityList>(dim);
			priorityList = new Vector<PriorityList>(dim);
			connection = new AtomicInteger();
			/* Popolo la struttura dati della priorita' con la classe 'testa'
			 * della lista */
			for (int i = blk.nextSetBit(0); i >= 0; i = blk.nextSetBit(i+1)){
				PriorityList pr = new PriorityList(i);
				blocksList.put(i, pr);
				priorityList.add(pr);
			}
			Client.log.out("[SWARM] Blocchi posseduti per il file '" +
				filecache.name()+"': "+filecache.getMyBlocks());
		}
		else{
			Client.log.out("[SWARM] Blocchi posseduti per il file '" +
				filecache.name()+"': Seeder.");
		}
	}
	
	private static SSLSocketFactory getSocketFactory() throws Exception{
		SSLContext ctx = SSLContext.getInstance("SSL");
		TrustManagerFactory tmf = TrustManagerFactory.getInstance(
				TrustManagerFactory.getDefaultAlgorithm());
		KeyStore trustStore = KeyStore.getInstance("JKS");
		char[] passphrase = "password".toCharArray();
		trustStore.load(new FileInputStream("bitcreek_client.jks"), passphrase);
		tmf.init(trustStore);
		ctx.init(null, tmf.getTrustManagers(), new SecureRandom()); 
		return ctx.getSocketFactory();
	}
	
	/**
	 * Recupera i Peer dello Swarm dal Tracker TCP, e popola lo swarm del
	 * client con la lista dei Peer ottenuta, inserendo solo quelli che e'
	 * riuscito a contattare.
	 * @return <code>True</code> se lo swarm del file e' stato ricevuto con
	 *         successo, <code>False</code> altrimenti. */
	protected boolean getSwarmFromServer(){
		SSLSocketFactory sslSkFactory = null;
		SSLSocket socket = null;
		ObjectOutputStream netOut = null;
		ObjectInputStream netIn = null;
		Message msg = null;
		boolean ret = false;
		PeerTalk peerinfo = null; 
		
		if (swarm!=null) swarm.clear();
		Client.log.out("[SWARM] Richiesto lo Swarm del file '"
				+filecache.name()+"'.");
		
		try{
			sslSkFactory = getSocketFactory();
			socket = (SSLSocket) sslSkFactory.createSocket(server,
				filecache.tor.getTrackerTCPPort());
			socket.setSoTimeout(Client.MAX_TIMEOUT);
			netOut = new ObjectOutputStream(socket.getOutputStream());
			netOut.writeObject(Message.swarmRequest(ClientImpl.myId,
				filecache.name()));
			netIn = new ObjectInputStream(socket.getInputStream());
			msg = (Message) netIn.readObject();
			switch(msg.code){
			case Message.SWARM_FOUND:
				Client.log.out("[SWARM] Lo swarm del file '"+
					filecache.name()+"' e' stato ricevuto con successo!");
				/* Se non sono un Seeder, inserisco i Peer ricevuti
				 * nello Swarm */
				if (!filecache.testSeeder() && msg.swarm!=null){
					Enumeration<Peer> p = msg.swarm.elements(); 
				    for (Peer peer: Collections.list(p)){
						/* Evito di inserire nello swarm me stesso */
						if (peer.idPeer.equals(ClientImpl.myId))
							continue;
						peerinfo = new PeerTalk(this, peer, nBlocchi);
				    	if (peerinfo.connect()){
				    		peerinfo.closeConnection();
				    		swarm.put(peerinfo.hashCode(), peerinfo);
				    		addInBlocksList(peerinfo);
				    	}
				    	else
				    		Client.log.err("[SWARM ERROR] Connessione con "
				    			+"il Peer "+peer.idPeer+" per il file '"
								+filecache.name()+ "' fallita.");
					}
					/* Ordina le informazioni ricevute */
					Collections.sort(priorityList);
					Client.log.out("[SWARM] Effettuato sort dello Swarm "
							+"del file '"+filecache.name()+"'.");
				}
				ret = true;
				break;
			case Message.SWARM_NOT_FOUND:
				Client.log.err("[SWARM ERROR] Swarm del descrittore '"
					+filecache.name()+"' non trovato!");		
				break;
			default:
				Client.log.err("[SWARM ERROR] Codice messaggio non " +
					"riconosciuto");
			}
		}
		catch (ClassNotFoundException e){
			Client.log.err("[SWARM ERROR] Messaggio non riconosciuto. | "
					+e.getMessage());
			Client.log.err(e);
		}
		catch (IOException e) {
			Client.log.err("[SWARM ERROR] Errore nel "
					+"contattare il Server. | "+e.getMessage());
		}
		catch (Exception e){		
			Client.log.err("[SWARM ERROR] Creazione Socket Factory.");				
			Client.log.err(e);
		}
		finally{
			try {
				if (socket!=null && !socket.isClosed())
					socket.close();
			}
			catch (IOException e){
				Client.log.err("[SWARM ERROR] Chiusura Socket SSL.");
				Client.log.err(e);
			}
		}
		return ret;
	}
	
	/**
	 * Inserisce un <code>Peer</code> nello <code>Swarm</code>.<br>
	 * L'inserimento di un nuovo Peer nello Swarm ha senso solo se il client
	 * non e' esso stesso un seeder, inoltre e' bene non inserire due volte
	 * lo stesso <code>Peer</code>.
	 * @param peer Il <code>Peer</code> che deve essere aggiunto allo
	 *         <code>Swarm</code>.
	 * @throws NullPointerException Se <code>InetSocketAddress</code> e'
	 *         <code>null</code>*/
	protected void addPeer(PeerTalk peer) throws NullPointerException{
		if (peer==null)	throw new NullPointerException(
				"[LISTENER] ERRORE Peer e' nullo!");
		swarm.put(peer.hashCode(), peer);
		addInBlocksList(peer);
		Collections.sort(priorityList);
		Client.log.out("[SWARM] Aggiunto nuovo Peer "+peer.getId()
				+" allo swarm del file '"+filecache.name()
				+"' ed effettuato sort.");
	}

	/**
	 * 
	 * @param peer Il <code>Peer</code> i cui blocchi devo essere aggiunti
	 *             alla lista di priorita' di questo <code>Swarm</code>.
	 * @throws NullPointerException Se il peer e' nullo.*/
	private synchronized void addInBlocksList(PeerTalk peer)
	throws NullPointerException{
		if (peer == null)
			throw new NullPointerException("Peer nullo! (addInBlocksList)");
		
		PriorityList peerList = null;
		if (peer.blocchi.isEmpty()) return;
		
		// Inserisco nella lista di priorita' solo i blocchi che non possiedo
		BitSet blocchi = (BitSet) filecache.getMyBlocks().clone();
		blocchi.flip(0, nBlocchi);
		blocchi.and(peer.blocchi);
			
		for (int i = blocchi.nextSetBit(0); i >= 0; i = blocchi.nextSetBit(i+1)){
			peerList = blocksList.get(i);
			if (peerList != null) peerList.count++;
		}
	}
	
	/**
	 * Aggiorna la lista dei blocchhi da scaricare in base al nuovo
	 * <code>BitSet</code> ricevuto dal Peer remoto, aggiungendo solo i 
	 * blocchi che ancora non si possegono e mantenendo aggiornato 
	 * l'ordinamento della lista. 
	 * @param blocchi Il <code>BitSet</code> di blocchi aggiornato ricevuto
	 *        dal Peer remoto*/ 
	protected synchronized void mergeBlocksList(BitSet blocchi){
		PriorityList peerList = null;
		for (int i = blocchi.nextSetBit(0); i >= 0; i = blocchi.nextSetBit(i+1)){
			if (!filecache.testMyBlock(i)){
				peerList = blocksList.get(i);
				if (peerList != null) peerList.count++;
			}
		}
		Client.log.out("[SWARM] Effettuato sort dello Swarm del file '"
				+filecache.name()+"'.");
		Collections.sort(priorityList);
	}
	
	/**
	 * Controlla l'esistenza del <code>Peer</code> nello <code>Swarm</code>.  
	 * @param idPeer L'<code>InetSocketAddress</code> del <code>Peer</code>
	 *        di cui si vuole confermare l'esistenza.
	 * @return <code>true</code> Se il <code>Peer</code> e' presente nello
	 *         <code>Swarm</code>, <code>false</code> altrimenti. */
	protected boolean existPeer(InetSocketAddress idPeer){
		return swarm.containsKey(idPeer.hashCode());
	}
	
	/**
	 * Ritorna la dimensione dello <code>Swarm</code>.
	 * @return La dimensione dello <code>Swarm</code>. */
	protected int getSwarmSize(){
		return swarm.size();
	}

	/**
	 * Questa funzione rimuove tutti i riferimenti ad un <code>Peer</code>
	 * dallo <code>Swarm</code> 
	 * @param peer Il <code>Peer</code> che si vuole rimuovere dallo
	 *        <code>Swarm</code>. */
	protected synchronized void removePeer(PeerTalk peer){
		PriorityList priority = null;
		/* Rimuove un Peer dallo Swarm */
		swarm.remove(peer.hashCode());
		/* Rimuove un Peer dalla lista di Priorita' */
		for (int i = peer.blocchi.nextSetBit(0); i >= 0; i = peer.blocchi.nextSetBit(i+1)){
			priority = blocksList.get(i);
			if (priority != null) priority.count--;
		}
		/* Ricalcola la priorita' di download dei blocchi */
		/* Ordinamento (se non sono un seeder) */
		Collections.sort(priorityList);
		Client.log.out("[SWARM] Rimosso Peer "+peer.getId()
				+" dallo swarm del file '"+filecache.name()
				+"' ed effettuato sort.");
	}
	
	/**
	 * Ritorna il blocco da scaricare a priorita' piu' alta che il
	 * <code>Peer</code> remoto possiede.<br> Il flag <i>download</i>
	 * del blocco viene temporaneamente settato a <code><b>TRUE</b></code>.
	 * <br> Se il blocco viene scaricato correttamente verra' rimosso dalla
	 * lista dei blocchi da scaricare,<br> altrimenti il flag <i>download</i>
	 * verra' settato nuovamente a <code><b>FALSE</b></code>. 
	 * @param blocchi Il <code>BitSet</code> dei blocchi posseduti dal 
	 *        <code>Peer</code> remoto.
	 * @return Un oggetto <code>PriorityList</code> che mantiene tutte le
	 *         informazioni sul blocco che si vuole scaricare. */
	protected synchronized PriorityList getRarestBlock(BitSet blocchi){
		Iterator<PriorityList> itr = priorityList.iterator();
		PriorityList pr = null;
		
		while (itr.hasNext()){
			pr = itr.next();
			if (blocchi.get(pr.indexBlock) && 
					pr.download.compareAndSet(false, true)){
				return pr;
			}
		}
		return null;
	} 
	
	/**
	 * Rimuove il blocco dalla lista dei blocchi da scaricare.
	 * @param block Il blocco che il <code>Peer</code> ha correttamente
	 *        scaricato ed ora viene rimosso dalla lista dei blocchi da
	 *        scaricare. */
	protected synchronized void removeBlock(PriorityList block){
		priorityList.remove(block);
		blocksList.remove(block.indexBlock);
	}

	/**
	 * Questo thread si preoccupera' di lanciare i threds per contattare i
	 * Peer dello Swarm che reperiranno i blocchi del file.<br> Se lo swarm
	 * si svuota contattera' nuovamente il Server per richiedere lo swarm
	 * aggiornato.<br> */
	public void run(){
		Enumeration<PeerTalk> peers = null;
		
		if (filecache.testSeeder())	return;
		Client.log.out("[SWARM] Thread per il file '"+filecache.name()
				+"' avviato.");

		while(!Client.shutdown && !shutdown){
			try{
				/* Se ci sono Peer dello swarm a cui non sono ancora connesso
				 * e non ho saturato il numero massimo di connessioni
				 * possibili... */
				if (connection.get() < swarm.size() && connection.get() < Client.MAX_CONN_OUT){			
					ArrayList<PeerTalk> peerlist =
						Collections.list(swarm.elements());
					Collections.sort(peerlist);
				    for (PeerTalk peer: peerlist){
				    	/* Client.log.out("[SWARM DEBUG]SIZE PEERLIST = "
				    			+ peerlist.size()+ " - "+ peer.getId()
				    			+" -C- "+ peer.blocchi.cardinality());	*/    	
				    	/* Se il Peer non e' gia' in uso */
				    	if (peer.inUse.compareAndSet(false, true)){
				    		if (peer.connect())
				    			thCommunications.execute(peer);
				    		else
				    			removePeer(peer);
				    	}
				    	/* Se ho raggiunto il numero massimo di connessioni
				    	 * in uscita termino il lancio dei threads */
				    	if (connection.get() >= Client.MAX_CONN_OUT) break;
				    }
				}
				
			    if (connection.get() >= Client.MAX_CONN_OUT){
			    	synchronized(connection){
					    try { connection.wait();}
					    catch (InterruptedException e1){
					    	Client.log.err(
					    		"[SWARM ERROR] Waiting on 'connection'.");
					    }
			    	}
			    }
			    else
			    	Thread.sleep(1000);
			    
				if (filecache.testSeeder()){
					shutdown = true;
					Client.log.out("[SWARM] Il file '"+filecache.name()
						+"' e' stato scaricato.");
					continue;
				}
				
				if (swarm.size()==0){
					Client.log.err("[SWARM] Non ci sono Peer per il file '"
							+filecache.name()+"'.");
					Client.log.out("[SWARM] Recupero nuovamente i Peer dello"
							+" swarm dal Server.");
					
					if (!getSwarmFromServer()) return;
					if (swarm.size()==0)
						Client.log.err("[SWARM] Il Server ha restituito"
							+" uno swarm vuoto per il file '"
							+filecache.name()+"'.");
				}
			}
			catch(Exception e){
				Client.log.err("[SWARM ERROR] "+e.getMessage());
				Client.log.err(e);
			}
		}
		
		thCommunications.shutdown();
		
		/* Arresto e chiudo tutte le connessioni in uscita con i peer
		 * remoti */
		peers = swarm.elements();
	    for (PeerTalk peer: Collections.list(peers)){
	    	peer.shutdown();
	    }
	    
		/* Attendo la chiusura dei threads dello swarm per lo scaricamento
		 * dei blocchi. */
		try {
			if (!thCommunications.awaitTermination(15, TimeUnit.SECONDS)){
				thCommunications.shutdownNow();
				if (!thCommunications.awaitTermination(15, TimeUnit.SECONDS)){
					Client.log.err("[SWARM ERROR] Il Pool di thread "
						+ "dello swarm '"+ filecache.name()
						+ "' non e' terminato correttamente.");
				}
			}
		}
		catch (InterruptedException ignore){
			thCommunications.shutdownNow();
			Thread.currentThread().interrupt();	
		}
		
		/* Elimino tutte le strutture dati dello swarm */
		blocksList.clear();		blocksList = null;
		priorityList.clear();	priorityList = null;
		swarm.clear();			swarm = null;
		filecache.flush();
		
		/* Se il Peer ha lasciato lo swarm chiudo il file e termino tutte
		 * le connessioni in ingresso da altri peer */
		if (!ClientImpl.swarmsList.containsKey(filecache.name())){
			Enumeration<MessageIn> msgIn = skPeers.elements();
		    for (MessageIn remotePeer: Collections.list(msgIn)){
		    	remotePeer.shutdown();
		    }			
		    /* APPUNTO: La terminazione del pool di thread delle
		     * comunicazioni ingresso non viene eseguita dato che i threads
		     * possono sempre essere usati dalla classe listener per altri
		     * swarm. */
			filecache.close();
		}
		
		Client.log.out("[SWARM] Thread per file '"+filecache.name()
				+"' arrestato.");
	}
	
	@Override
	protected void finalize() throws Throwable{
		filecache.close();
		super.finalize();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "ClientSwarm [Swarm Name = "+filecache.name()
				+"\nBlocchi posseduti = " + filecache.getMyBlocks()
				+"\nNumero blocchi del file = " + nBlocchi 
				+"\nNumero di connessioni aperte = " + connection.get()
				+"\nPriority List Size = "+priorityList.size()
				+"\nBlocks List Size = "+blocksList.size()
				+"\nNumero di Peer nello Swarm = " + swarm.size()
				+"]";
	}
}
