package bitcreek.client;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.security.NoSuchAlgorithmException;
import java.util.BitSet;
import java.util.Set;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;

import bitcreek.Torrent;

/**
 * Questa classe si occupa di eseguire la scrittura/lettura dei blocchi
 * del file che si sta scaricando/inviando e di implementare una cache
 * per i blocchi stessi, al fine di ottimizzare l'accesso al disco e il
 * reperimento dei blocchi.<br><br>
 * 
 * Aumentando la dimensione della cache si avranno migliori prestazioni,
 * ma di contro ci sara' uno spreco maggiore di memoria.<br>
 * 
 * Per ogni blocco scaricato verra' calcolato il checksum e confrontato con
 * quello memorizzato nel descrittore del file, solo se i due valori
 * coincidono il blocco viene ritenuto essere stato scaricato correttamente.
 *
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732<br>
 * 	     	Progetto LPR anno 2008/2009. */
public class FileCache {
	/** Numero massimo di blocchi mentenuti in memoria */
	private long pending_size = 1000;
	/** Se e' impostata a true la scrittura sul disco e' immediata.
	 * (no cache) */
	private boolean	direct_writes = false;
	/** Cache dei blocchi da scrivere su file */
	private ConcurrentHashMap<Integer, byte[]> pendingWrite = null;
	/** Mantiene lo status dei blocchi scaricati dal Peer */
	private BitSet bs_blocks = null;
	/** Nome di default della cartella dove verranno scaricati i file */
	private static final String	DOWNLOAD_DIR_NAME = "download";
	/** Cartella dove verranno scaricati i file */
	private File download_dir = null;
	/** Puntatore al file */
	private RandomAccessFile file = null;
	/** Descrittore del file*/
	protected final Torrent tor;
	/** Numero di blocchi del file */
	private final int numBlocchi;
	
	/**
	 * Costruttore privato della classe
	 * @param tor Il torrent che si sta scaricando/uppando
	 * @param locale Se impostato a <code><b>TRUE</b></code> indica che il
	 *        file e' stato creato da questo client e per questo motivo il
	 *        percorso del file e' stato definito dall'utente. Altrimenti, 
	 *        se impostato a <code><b>FALSE</b></code> il percorso dove il 
	 *        file verra' memorizzato e' quello di default del programma.
	 * @param pending_size Numero massimo di blocchi mentenuti in memoria
	 * @param direct_writes Se e' impostata a true la scrittura sul disco e'
	 *        immediata.
	 * @throws IOException 
	 * @throws NoSuchAlgorithmException */
	private FileCache(Torrent tor, boolean locale, long pending_size, boolean direct_writes)
	throws IOException,	NoSuchAlgorithmException{
		this.pending_size = pending_size;
		this.direct_writes = direct_writes;
		numBlocchi = tor.numBlocchi();
		bs_blocks = new BitSet(numBlocchi);
		createDownloadDir(DOWNLOAD_DIR_NAME);
		this.tor = tor;
		File f = null;
		
		if (!direct_writes)
			pendingWrite = new ConcurrentHashMap<Integer, byte[]>();
			
		if (locale){
			f = new File(tor.percorso);	
			if (!f.exists())
				throw new IOException("File torrent non trovato");
			bs_blocks.flip(0, tor.numBlocchi());
		}
		else{
			f = new File(download_dir, tor.name);
			if (!f.exists()) f.createNewFile();
		}
		file = new RandomAccessFile(f,"rwd");
		if (!locale){
			if (file.length()!=tor.length)
				/* Alloco un file della dimensione
				 * del file che devo scaricare */
				file.setLength(tor.length);
			else
				checkFile();
		}
	}
	
	/**
	 * File Torrent creato dal Client che ha pubblicato il file
	 * @param tor Il torrent che si sta uppando 
	 * @return Ritorna un istanza della classe
	 * @param pending_size Numero massimo di blocchi mentenuti in memoria
	 * @param direct_writes Se e' impostata a true la scrittura sul disco e'
	 *        immediata. 
	 * @throws IOException 
	 * @throws NoSuchAlgorithmException */
	public static FileCache createFileCacheFromRemoteTorrent(Torrent tor, long pending_size, boolean direct_writes)
	throws NoSuchAlgorithmException, IOException{	
		return new FileCache(tor,false,pending_size,direct_writes);
	}

	/** 
	 * File Torrent scaricato dal Server
	 * @param tor Il torrent che si sta scaricando/uppando
	 * @return Ritorna un istanza della classe
	 * @param pending_size Numero massimo di blocchi mentenuti in memoria
	 * @param direct_writes Se e' impostata a true la scrittura sul disco e'
	 *        immediata. 
	 * @throws IOException 
	 * @throws NoSuchAlgorithmException */
	public static FileCache createFileCacheFromLocalTorrent(Torrent tor, long pending_size, boolean direct_writes)
	throws NoSuchAlgorithmException, IOException{
		return new FileCache(tor,true,pending_size,direct_writes);
	}

	/**
	 * Controlla quali blocchi sono stati gia' scaricati. 
	 * @throws IOException 
	 * @throws NoSuchAlgorithmException */
	private void checkFile() throws IOException, NoSuchAlgorithmException{
		if (file.length()==0) return;
		
	    int offset=0, numRead=0;
		long length = file.length();
		/* Calcolo il numero dei blocchi */
		int blocks =(int)(length/Torrent.blockSize);
	    if ((length%Torrent.blockSize)!=0) blocks++;
			
	    for (int i=0; offset < length; i++){	
	    	byte[] blocco = new byte[Torrent.blockSize];
	       	numRead=file.read(blocco, 0, Torrent.blockSize);   	
	       	if (tor.checkBlock(i, blocco)) 
	       		bs_blocks.set(i);
		   	offset += numRead;
	    }
	}
	
	/** Controlla l'esistenza della cartella di download 
	 * @param debug_dir_name Nome della cartella dove verranno scaricati i
	 *                       file. */
	private void createDownloadDir(String debug_dir_name){
		File file = new File(debug_dir_name);
		if ((file.exists() && file.isFile())){
			file.delete();
		}
		if (!file.exists()){
			file.mkdir();
		}
		download_dir = file;
	}
	
	/**
	 * Aggiunge alla cache, se attiva, ed eventulamente, se e' piena,
	 * scrive su file i blocchi che sono presenti nella cache.
	 * Se la cache non e' attiva la scrittura su file e' immediata.
	 * Memorizza il possesso del blocco mettendo il corrispettivo indice
	 * del BitSet 'bs_blocks' a True.
	 * @param blockIndex Indice del blocco
	 * @param blocco Il blocco scaricato
	 * @return True se il blocco e' stato inserito con successo */
	public synchronized boolean put(int blockIndex, byte[] blocco){
		try{
			if (!tor.checkBlock(blockIndex, blocco)){
				Client.log.err("[FILECACHE ERROR] Blocco '"+blockIndex+
						"', del file '"+tor.name+ "' SHA1 errato!");				
				return false;
			}
			/* Contolla se si possiede gia' il blocco */
			if (bs_blocks.get(blockIndex)){
				Client.log.err("[FILECACHE ERROR] Blocco '"+blockIndex+
					"', del file '"+tor.name+ "', gia' presente!");
				return false;
			}
			
			/* Imposta il blocco come posseduto */
			bs_blocks.set(blockIndex);
			
			if (!direct_writes){
				pendingWrite.put(blockIndex, blocco);
				if (pendingWrite.size() >= pending_size)
					writePending();
			}
			else
				writeNow(blockIndex, blocco);
			return true;
		}
		catch (IndexOutOfBoundsException e){
			Client.log.err("[FILECACHE ERROR] Indice " +blockIndex
					+ " fuori range!");
			Client.log.err(e);
		}
		catch (Exception e){Client.log.err(e);}
		return false;
	}
	
	/** 
	 * Cerca il blocco nella cache, se attiva, e lo restituisce
	 * se presente, altrimenti accede al file posto sul disco  
	 * @param blockIndex L'indice del blocco richiesto
	 * @return Il blocco richiesto */
	public byte[] get(int blockIndex){
		byte[] blocco = null;
		try{
			/* Controlla se il blocco richiesto e' in un range corretto */
			if (blockIndex>=bs_blocks.size() || blockIndex<0)
				throw new IndexOutOfBoundsException(" Errore blockIndex");
			
			/* Controlla se si possiede il blocco richiesto */
			if (!bs_blocks.get(blockIndex)){
				Client.log.err("[FILECACHE ERROR] Non si possiede il"
						+ " blocco (" + blockIndex + ") richesto !");
				return null;
			}
			if (!direct_writes)
				blocco = pendingWrite.get(blockIndex);
			
			/* Se il blocco non e' presente nella cache 
			 * o se 'direct_writes' e' true... */
			if (blocco == null)
				blocco = read(blockIndex);
		}
		catch(IndexOutOfBoundsException e){
			Client.log.err("[FILECACHE ERROR] Indice ("+blockIndex
					+") fuori range!");
			Client.log.err(e);
		}
		catch (FileNotFoundException e){Client.log.err(e);}
		catch (IOException e){Client.log.err(e);}
		return blocco;
	}
	
	/** Azzera la cache scrivendo i blocchi pendenti sul disco.*/
	public synchronized void flush(){
		if (!direct_writes){
			try {writePending();}
			catch (IOException e){Client.log.err(e);}
		}
	}
	
	/**
	 * Azzera la cache dei blocchi scrivendone i pendenti sul disco e
	 * chiude il file.<br>
	 * Quando un client esce dallo <code>Swarm</code> o il programma e'
	 * in fase fi chiusura si <b>DEVE</b> richiamare questa funzione! */
	public synchronized void close(){
		if (!direct_writes){
			try {writePending();}
			catch (IOException e){Client.log.err(e);}
		}
		try {file.close();}
		catch (IOException e){Client.log.err(e);}
	}
	
	/** Srittura sul disco dei blochi pendenti della cache 
	 * @throws IOException */
	private void writePending() throws IOException{
		if (pendingWrite == null) return;
		Set<Entry<Integer, byte[]>> to_write = pendingWrite.entrySet();
		for (Entry<Integer, byte[]> entry : to_write){
			file.seek(entry.getKey()*Torrent.blockSize);
			file.write(entry.getValue());
		}
		pendingWrite.clear();
	}
	
	/** Scrittura sul disco immediata del blocco ricevuto 
	 * @param blockIndex Indice del blocco che si sta scrivendo su file
	 * @param blocco Dati del blocco che si stannno scrivendo su file
	 * @throws IOException */
	private void writeNow(int blockIndex, byte[] blocco)
	throws IOException{
		file.seek(blockIndex*Torrent.blockSize);
		file.write(blocco);
	}
	
	/**
	 * Legge da file il blocco richiesto 
	 * @param blockIndex Indice del blocco richiesto
	 * @return Il blocco richiesto
	 * @throws FileNotFoundException Se non viene trovato il file sul disco
	 * @throws IOException */
	private byte[] read(int blockIndex)
	throws FileNotFoundException, IOException {
		
		/* Si posiziona sul blocco da leggere */
		try{ file.seek(Torrent.blockSize*blockIndex); }
		catch (IOException e){
			Client.log.err("[LISTENER ERROR] Lettura blocco errata! " +
					"E' stato cancellato il file dal disco?");
			Client.log.err(e);
			return null;
		}
		byte blocco[] = new byte[Torrent.blockSize];
		/* Legge il blocco dal disco */
		file.read(blocco, 0, Torrent.blockSize);
		/* Domanda: E' opportuno controllare il digest del blocco letto? */
		return blocco;
	}
	
	/**
	 * Ritorna lo status dei blocchi scaricati dal Peer. 
	 * @return Ritorna BitSet che mantiene lo status dei blocchi scaricati
	 * 		   dal Peer */
	public synchronized BitSet getMyBlocks(){
		return bs_blocks;
	}
	
	/**
	 * Ritorna True se si possiede il blocco
	 * @param blockIndex L'indice del blocco di cui si vuole accertarsi il
	 * 		  possesso.
	 * @return True se si possiede il blocco */
	public synchronized boolean testMyBlock(int blockIndex){ 
		return bs_blocks.get(blockIndex);
	}
	
	/**
	 * Ritorna True se si possiedono tutti i blocchi
	 * @return Ritorna True se si possiedono tutti i blocchi */
	public synchronized boolean testSeeder(){ 
		return (bs_blocks.cardinality() == tor.numBlocchi()); 
	}
	
	/** Funzione per il solo scopo di debug del codice */
	public synchronized void debug(){
		if (bs_blocks==null)
			System.err.println("bs_blocks e' nullo");
		else{
			System.err.println("Lunghezza = "+bs_blocks.length());
			System.err.println("Cardinalita' = "+bs_blocks.cardinality());
			System.err.println(bs_blocks);
		}
	}
	
	/**
	 * Restituisce il nome del file 
	 * @return Il nome del File. */
	public String name(){ return tor.name; }
	
	/**
	 * Restituisce il numero di blocchi del file 
	 * @return Il numero di blocchi del file. */
	public int getNumBlocchi(){return numBlocchi;}
	
	@Override
	protected void finalize() throws Throwable{
		close();
		super.finalize();
	}
}
