package bitcreek.client;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.util.Collection;
import java.util.Iterator;

import bitcreek.Message;

/**
 * Questa classe si preoccupa di inviare ad intervalli regolari di tempo un
 * messaggio di KeepAlive, tramite l'uso di una socket udp, al Server per
 * notificare la propria presenza nello swarm a cui il client si e'
 * registrato.<br>
 * Viene mandato un messaggio di KeepAlive per ogni swarm a cui il Peer
 * appartiene; il mancato invio di uno di questi messaggi per un periodo di
 * tempo superiore a un tempo massimo definito nel Server provoca l'uscita
 * del client da quel determinato swarm.<br>
 * Questo e' l'unico metodo, che il client ha per potersi cancellare da uno
 * swarm. L'intervallo di tempo e' definito nella classe Client del
 * programma attraverso l'attributo statico </code>keepAliveTime</code> che
 * puo' essere settato a tempo di esecuzione attraverso il comando
 * <code>-Dkeepalive=n</code> dove n e' un intero che esprime l'intervallo
 * di tempo in millisecondi tra un keepalive ed un altro.<br>
 * Altrimenti se non diversamente specificato verra' usato il tempo di
 * default definito dall'attributo statico <code>KEEPALIVE_DEF</code>
 * presente nella classe Client di questo programma.
 * In ogni messaggio mandato al Server il Peer indica l'indirizzo e la
 * porta TCP su cui e' in ascolto, che lo identificano in modo univoco e il
 * nome del file, che identifica univocamente lo swarm. 
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732<br>
 * 	     	Progetto LPR anno 2008/2009. */
public class KeepAlive implements Runnable {
	/** Dimensione massima dei dati del messaggio */
	public static final int MAX_UDP_PAYLOAD = 1024;
	/** */
	public static int BUFFER_SIZE = 1024;
	/** Socket di ricezione del tracker udp */
	private DatagramSocket sk_send = null;
	/** Pacchetto Datagram che invio */
	private DatagramPacket dp_send = null;
	private InetAddress remoteServer = null;
	
	/** Costruttore di classe
	 * @param remoteServer L'indirizzo del <code>Peer</code> con cui sto
	 *        comunicando.
	 * @throws SocketException */
	public KeepAlive(InetAddress remoteServer) throws SocketException{
		this.remoteServer = remoteServer;
		if (ClientImpl.myId==null){
			Client.log.err("[KEEPALIVE ERROR] 'myId'" +
				" del client non e' settato!");
			throw new NullPointerException();
		}
		this.sk_send = new DatagramSocket();	/* Socket d'invio */
	}

	/** Restituisce un pacchetto contenente l'oggetto passato per
	 * 	argomento, serializzato, nel payload.
	 * @param obj
	 * @return Restituisce un pacchetto contenente l'oggetto passato per
	 *  	   argomento, serializzato, nel payload.
	 * @throws IOException
	 */
	public static DatagramPacket obj2dp(Object obj) throws IOException{
		ByteArrayOutputStream bos = new ByteArrayOutputStream(); 
		ObjectOutputStream object_output_stream =
				new ObjectOutputStream(bos);
		object_output_stream.writeObject(obj);
		byte []data = bos.toByteArray();
		DatagramPacket dp = new DatagramPacket(data, data.length);
		return dp;			
	}
	
	public void run(){
		Client.log.out("[KEEPALIVE] Thread avviato.");
		while(!Client.shutdown){
			if (ClientImpl.swarmsList!=null){
				Collection<ClientSwarm> c = ClientImpl.swarmsList.values();
				Iterator<ClientSwarm> itr = c.iterator();
				while (itr.hasNext()){
					try {
						ClientSwarm swarm = itr.next();
						dp_send = obj2dp(Message.keepalive(ClientImpl.myId,
								swarm.filecache.name()));
						dp_send.setAddress(remoteServer);
						dp_send.setPort(
								swarm.filecache.tor.getTrackerUDPPort());
						sk_send.send(dp_send);					
						Client.log.out("[KEEPALIVE] Per il file '"
							+swarm.filecache.name()
							+"' inviato al server.");
					}
					catch (IOException e){
						Client.log.err("[KEEPALIVE ERROR] Socket. "
							+e.toString());
						Client.log.err(e);
					}
				}
			}
			try {Thread.sleep(Client.keepAliveTime);}
			catch (InterruptedException e){
				if (!Client.shutdown){
					Client.log.err("[KEEPALIVE ERROR] Socket. "
							+e.getMessage());
					Client.log.err(e);
				}
			}
		}
		Client.log.out("[KEEPALIVE] Thread arrestato.");
	}
}
