package bitcreek.client;

import java.io.EOFException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketException;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import bitcreek.Message;
import bitcreek.Peer;

/**
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732<br>
 * 	     	Progetto LPR anno 2008/2009. */
public class Listener implements Runnable {
	
	/** Pool di thread per le comunicazioni in ingresso  */
	protected static ThreadPoolExecutor thListener =
		new ThreadPoolExecutor(10, 200, 50000L,
			TimeUnit.MILLISECONDS, new LinkedBlockingQueue<Runnable>());
	
	private static final int TIMEOUT_REPLY = 60000; /* 60 secondi */
	
	/** Arresta il thread che accetta le connessioni da altri client */
	public void shutdown(){
		Client.log.out("[LISTENER] Arresto Thread in corso...");
		try{ ClientImpl.socketTCP.close(); }
		catch (IOException e){Client.log.err(e);}
	}

	public void run() {
		Client.log.out("[LISTENER] Thread avviato.");
		Socket skClient = null;
		while(!Client.shutdown){
			try {
				skClient = ClientImpl.socketTCP.accept();
				thListener.execute(new MessageIn(skClient));
			}
			catch (IOException e){
				if (!Client.shutdown){
					Client.log.err("[LISTENER ERROR] "+e.getMessage());
					Client.log.err(e);
				}
			}
		}
		
		thListener.shutdown();
		/* Attendo la chiusura dei threads del listener */
		try {
			if (!thListener.awaitTermination(15, TimeUnit.SECONDS)){
				thListener.shutdownNow();
				if (!thListener.awaitTermination(15, TimeUnit.SECONDS)){
					Client.log.err("[LISTENER ERROR] Il Pool di thread" +
							" non e' terminato correttamente.");
				}
			}
		}
		catch (InterruptedException ignore){
			thListener.shutdownNow();
			Thread.currentThread().interrupt();	
		}
	}
	
	private enum Stato {
		CONNECTED,
		CLOSED,
		REMOTE_CLOSED,
	}
	
/* ______________________________________________________________________ */
	
	protected class MessageIn implements Runnable{
		private volatile Stato status = Stato.CONNECTED;
		private InetSocketAddress peerId = null;
		private Socket skClient = null;
		private ObjectInputStream netIn = null;
		private ObjectOutputStream netOut = null;
		private String filename = null;
		private Message msg = null;
		private ClientSwarm swarm = null;
		
		/**
		 * Costruttore.
		 * @param skClient La socket di comunicazione */
		public MessageIn(Socket skClient){
			this.skClient = skClient;
		}
		
		/** Arresta il thread che accetta le connessioni da altri client */
		public void shutdown(){
			Client.log.out("[LISTENER] Reply Thread arresto in corso...");
			try{
				netIn.close();
				netOut.close();
				skClient.close();
			}
			catch (IOException e){/*Client.log.err(e);*/}
		}
		
		public void run(){
			Client.log.out("[LISTENER] Accettata nuova richiesta di" +
					" connessione.");
			try{
				netIn = new ObjectInputStream(skClient.getInputStream());
				netOut = new ObjectOutputStream(skClient.getOutputStream());
				msg = (Message) netIn.readObject();
				switch(msg.code){
				case Message.CONN_REQ:
					filename = msg.filename;
					swarm = ClientImpl.swarmsList.get(filename);
					peerId = msg.idPeer;
					if (swarm==null){
						Client.log.err("[LISTENER ERROR] Connessione con il"
							+" Peer "+peerId+" rifiutata.\nQuesto Peer non "
							+"sta scaricando il file '"+filename+"'.");
						netOut.writeObject(Message.swarmNotFound());
						netOut.writeObject(Message.connErr(ClientImpl.myId));
						status = Stato.CLOSED;
						closeConnection();
						return;
					}
						
					synchronized(this){
						/* Controllo che non abbia gia' raggiunto il numero
						 * massimo di connessioni per questo swarm */ 
						if (swarm.skPeers.size()>=Client.MAX_CONN_IN){
							Client.log.err("[LISTENER ERROR] Connessione"
								+" con il Peer "+peerId+" rifiutata.\n"
								+"Numero massimo di connessioni in ingresso"
								+" per lo swarm '"+filename+"' raggiunto.");
							netOut.writeObject(
									Message.connErr(ClientImpl.myId));
							status = Stato.CLOSED;
							closeConnection();
							return;						
						}
						
						/* Controllo se il Peer e' gia' connesso */
						if (swarm.skPeers.containsKey(peerId.hashCode())){
							Client.log.err("[LISTENER ERROR] Connessione "
								+"con il Peer "+peerId
								+" rifiutata. Peer gia' connesso.");
							netOut.writeObject(
									Message.connErr(ClientImpl.myId));
							status = Stato.CLOSED;
							closeConnection();
							return;
						}
						netOut.writeObject(Message.connAck(ClientImpl.myId,
								swarm.filecache.getMyBlocks()));
						swarm.skPeers.put(peerId.hashCode(), this);
						if (!swarm.filecache.testSeeder() && !swarm.existPeer(peerId)){
							swarm.addPeer(new PeerTalk(swarm, new Peer(peerId), msg.bs_blocchi));
						}	
					}
					Client.log.out("[LISTENER] Connessione con il Peer "
						+peerId+" accettata. - File '"+filename+"'. <--");
					break;
				default:
					Client.log.err("[LISTENER ERROR] Messaggio '"+msg.code+
							"' non previsto. (reply)");
					status = Stato.CLOSED;
					closeConnection();
					return;
				}
			}
			catch (ClassNotFoundException e){
				Client.log.err("[LISTENER ERROR]" +
						" Oggetto messaggio non riconosciuto.");
				Client.log.err(e);
				status = Stato.CLOSED;
				closeConnection();
				return;
			}
			catch (IOException e){
				Client.log.err("[LISTENER ERROR] Errore di comunicazione" 
						+" con il Peer "+peerId+" per il file '"+filename
						+"'. | (HandShake)");
				Client.log.err(e);
				status = Stato.CLOSED;
				closeConnection();
				return;
			}
			/* FINE HANDSHAKE */
			
			while (status == Stato.CONNECTED && !Client.shutdown){
				reply();
			}
			closeConnection();
			swarm.skPeers.remove(peerId.hashCode());
			return;
		}
		
		private void reply(){
			try{
				skClient.setSoTimeout(TIMEOUT_REPLY);
				msg = (Message) netIn.readObject();
				switch(msg.code){
				case Message.BLOCK_REQ:
					byte[] blocco = swarm.filecache.get(msg.indice);
					if (blocco!=null){
						netOut.writeObject(Message.blockSend(msg.indice,blocco));
						Client.log.out("[LISTENER] Inviato blocco "
							+msg.indice+" del file '" +filename+"' al Peer "
							+peerId);
					}	
					else{
						netOut.writeObject(Message.blockErr(msg.indice));
						Client.log.err("[LISTENER ERROR] Blocco "+msg.indice
							+" del file '"+filename+"' richiesto dal Peer "
							+peerId+" non posseduto.");
					}
					break;
				case Message.BITSET_GET: 
   /** 
	* NON CANCELLARE QUESTO COMANDO, se non si vuole che la JVM faccia casini!!  (O_O)'
	* @see http://java.sun.com/javase/technologies/core/basic/serializationFAQ.jsp#handle */
					netOut.reset();	
					Client.log.out("[LISTENER] Richiesto dal Peer "+ peerId 
							+" il BitSet dei blocchi <---");
					netOut.writeObject(
						Message.sendPeerBlocks(swarm.filecache.getMyBlocks()));					
					Client.log.log("[LISTENER] Inviato il BitSet dei blocchi"
							+" al Peer "+ peerId+" --->");
					break;
				default:
					Client.log.err("[LISTENER ERROR] Messaggio '"+msg.code+
							"' non previsto. (reply)");
					status = Stato.CLOSED;
					break;
				}
			}
			catch (SocketException e){
				Client.log.out("[LISTENER] Connessione remota chiusa"
						+ " dal Peer "+peerId+" - File '"+filename+"'. "
						+ e.getMessage());
				status = Stato.REMOTE_CLOSED;
			}
			catch (EOFException e){
				Client.log.out("[LISTENER] Connessione remota chiusa"
						+" dal Peer "+peerId+" - File '"+filename+"'.");
				status = Stato.REMOTE_CLOSED;
			}
			catch (ClassNotFoundException e){
				Client.log.err("[LISTENER ERROR]" +
						" Oggetto messaggio non riconosciuto.");
				Client.log.err(e);
				status = Stato.CLOSED;
			}
			catch (IOException e){
				Client.log.err("[LISTENER ERROR] Errore di comunicazione"
						+" con il Peer "+peerId+" per il file '"+filename+
						"'. (reply)");
				Client.log.err(e);
				status = Stato.CLOSED;
			}
		}
		
		private void closeConnection(){
			try{
				if (!skClient.isClosed()) skClient.close();
				if (status != Stato.REMOTE_CLOSED){
					if (filename!=null && peerId!=null){
						Client.log.out("[LISTENER] Connessione remota"
							+" chiusa con il Peer "+peerId+" - File '"
							+filename+"'.");
					}
					else{
						Client.log.out("[IO ERROR]"
							+" Connessione remota chiusa.");
					}
				}
			}
			catch (IOException e){Client.log.err(e);}
			Client.log.out("[LISTENER] Reply Thread arrestato.");
		}
	}
}
