package bitcreek.client;

import java.io.EOFException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.ConnectException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.util.BitSet;
import java.util.concurrent.atomic.AtomicBoolean;

import bitcreek.Message;
import bitcreek.Peer;
import bitcreek.client.ClientSwarm.PriorityList;

/**
 * Mantiene le informazioni sul Peer remoto.
 * Crea un canale di comunicazione verso il peer remoto a cui quest'oggetto
 * e' associato ne scarica i blocchi posseduti.
 * Durante la procedura iniziale di Handshake il Peer locale fornisce, le
 * informazioni sulla sua identita' e su i blocchi posseduti, al Peer
 * remoto che rispondera' inviando a sua volta il BitSet dei blocchi che 
 * possiede.
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732
 * 	     	Progetto LPR anno 2008/2009. */
public class PeerTalk implements Runnable, Comparable<Object>{
	
	private enum Stato {
		CLOSED,
		CONNECTED,
		REMOTELY_CLOSED,
		ERROR
	}
	
	private final Peer peer;
	/** Il <code>BitSet</code> blocchi posseduti da un <code>Peer</code> */
	protected BitSet blocchi;
	/** Lo swarm a cui il <code>Peer</code> appartiene */
	private ClientSwarm swarm;
	/**
	 * Se impostata a TRUE arresta, se in esecuzione, il thread di questa
	 * classe */
	private volatile boolean shutdown = false;

	/* Variabili per la comunicazione */
	/** Mantiene traccia dello stato della connessione. */
	private volatile Stato status = Stato.CLOSED;
	private String filename;
	private Socket skClient = null;
	private ObjectInputStream netIn = null;
	private ObjectOutputStream netOut = null;
	private Message msg = null;
	private static final int TIMEOUT = 60000;
	protected AtomicBoolean inUse = new AtomicBoolean();
	
	/**
	 * Costruttore Crea un nuovo Peer, settando la dimensione del BitSet
	 * Questo costruttore viene utilizzato quando si riceve lo Swarm dal
	 * server, che non possiede nessuna informazione su i blocchi posseduti
	 * da i Peer.  
	 * @param swarm Lo Swarm a cui il <code>Peer</code> appartiene
	 * @param peer Il Peer di cui si vogliono immgazzinare le informazioni.
	 * @param nBlocchi La dimensione del <code>BitSet</code> dei blocchi
	 *        del Peer.
	 * @throws NullPointerException Se il<code>Peer</code> e'
	 *        <code>null</code> o se lo swarm e' <code>null</code> .*/
	public PeerTalk(ClientSwarm swarm, Peer peer, int nBlocchi)
	throws NullPointerException{
		if (peer==null) throw new NullPointerException(
				"Il Peer non puo' essere nullo!");
		if (swarm==null) throw new NullPointerException(
				"Lo Swarm non puo' essere nullo!");
		
		this.swarm = swarm;
		this.peer = peer;
		filename = swarm.filecache.name();
		blocchi = new BitSet(nBlocchi);
	}
	
	/**
	 * Costruttore specifico: Crea un Peer e inizializza il BitSet
	 * dei blocchi posseduti.
	 * Questo costruttore viene utilizzato quando un Peer non presente nello
	 * swarm locale di questo client lo contatta. Il Peer viene aggiunto
	 * nello swarm locale del client settando immediatamente il BitSet dei 
	 * blocchi che possiede.
	 * @param swarm Lo Swarm a cui il <code>Peer</code> appartiene
	 * @param peer Il Peer di cui si vogliono immgazzinare le informazioni.
	 * @param blocchi Il <code>BitSet</code> stesso dei blocchi del Peer.
	 * @throws NullPointerException Se il<code>Peer</code> o il 
	 *         <code>BitSet</code> dei blocchi e' </code>null</code>.*/
	public PeerTalk(ClientSwarm swarm, Peer peer, BitSet blocchi)
	throws NullPointerException{
		if (peer==null) throw new NullPointerException(
				"Il Peer non puo' essere nullo!");
		if (blocchi==null) throw new NullPointerException(
				"IL BitSet 'blocchi' non puo' essere nullo!");
		if (swarm==null) throw new NullPointerException(
				"Lo swarm deve essere inizializzato correttamente!");
		this.peer = peer;
		this.blocchi = blocchi;
		this.swarm = swarm;
		filename = swarm.filecache.name();
	}
	
	@Override
	public int hashCode(){return peer.hash;}

	@Override
	public boolean equals(Object obj){
		return peer.idPeer.equals(obj);
	}
	
	/* Il primo Peer della lista dello swarm sara' il Peer con il maggior
	 * numero di blocchi in suo possesso */
	public int compareTo(Object obj){
		PeerTalk peer = (PeerTalk) obj;
		int size = peer.blocchi.cardinality();
		return (blocchi.cardinality()>size) ? -1 : (blocchi.cardinality()<size) ? 1 : 0;
	}

	/**
	 * Restitisco al chiamante l'indirizzo <code>InetAddress</code>
	 * e la porta che identifica il <code>Peer</code>.
	 * @return InetSocketAddress */
	public InetSocketAddress getId(){
		return peer.idPeer;
	}
	
	/** Forza l'arresto del thread */
	protected void shutdown(){
		shutdown = true;
		status = Stato.CLOSED;
		if (skClient!=null){ 
			try {
				netIn.close();
				netOut.close();
				skClient.close();
			}
			catch (IOException e){Client.log.err(e);}
		}
	}
	
	/*___________________________________________________________________
	 * 
	 * 						FUNZIONI DI COMUNICAZIONE
	 *___________________________________________________________________
	 * */
	/**
	 * Si collega al <code>Peer</code> remoto ed effettua l'handshake
	 * scambiando il proprio BitSet con il <code>Peer</code> remoto e 
	 * viceversa.
	 * @return <code>TRUE</code> se si e' connesso correttamente al
	 *         <code>Peer</code> remoto, <code>FALSE</code> altrienti.*/
	protected synchronized boolean connect(){
		try {
			skClient = new Socket();
			skClient.connect(peer.idPeer, 0);
			netOut = new ObjectOutputStream(skClient.getOutputStream());
			netIn = new ObjectInputStream(skClient.getInputStream());
			msg = null;
			Client.log.out("[TALKER] Connessione in corso con il Peer "
					+peer.idPeer+". File: '"+filename+"'. --->");
			skClient.setSoTimeout(TIMEOUT);
			netOut.writeObject(Message.connReq(ClientImpl.myId,
				filename, swarm.filecache.getMyBlocks()));
			msg = (Message) netIn.readObject();
			switch(msg.code){
			case Message.CONN_ACK:
				blocchi = msg.bs_blocchi;
				status = Stato.CONNECTED;
				/* Incremento il conteggio delle connessioni aperte */
				swarm.connection.incrementAndGet();
				Client.log.out("[TALKER] Connessione con il Peer "
						+peer.idPeer+" effettuata. File: '"+filename
						+"'. <---");
				break;
			case Message.CONN_ERR:
				Client.log.err("[TALKER ERROR] Il Peer " + peer.idPeer +
					" ha rifiutato la connessione.");
				status = Stato.ERROR;
				break;
			case Message.SWARM_NOT_FOUND:
				Client.log.err("[TALKER ERROR] Il Peer contattato "
						+peer.idPeer+" non sta scaricando il file '"
						+filename+"'.");
				status = Stato.ERROR;
				break;				
			default:
				Client.log.err("[TALKER ERROR] Messaggio '"+msg.code+
						"' non previsto. (reply)");
				status = Stato.ERROR;
				break;				
			}
		}
		catch (ClassNotFoundException e){
			Client.log.err("[TALKER ERROR] Oggetto messaggio" +
					" non riconosciuto.");
			Client.log.err(e);
			status = Stato.ERROR;
		}		
		catch (SocketTimeoutException e){
			Client.log.err("[TALKER ERROR] Errore connessione:" +
					" SocketTimeoutException.");
			status = Stato.ERROR;
		}
		catch (ConnectException e){
			Client.log.err("[TALKER ERROR] Errore nella connessione con "+
				"il Peer '"+peer.idPeer+"'. "+e.toString());
			status = Stato.ERROR;
		}	
		catch (EOFException e){
			Client.log.err("[TALKER] Il Peer "+peer.idPeer+
				"' ha rifiutato la richiesta di connessione.");
			status = Stato.ERROR;
		}
		catch (IOException e) {
			Client.log.err("[TALKER ERROR] Impossibile stabilire un canale "
					+"di comunicatione verso il Peer "+peer.idPeer+".");
			Client.log.err(e);
			status = Stato.ERROR;
		}
		catch (NullPointerException e){
			Client.log.err("[TALKER ERROR] Errore nello stabilire una "+
					"comunicazione verso il Peer "+peer.idPeer+".");
			Client.log.err(e);
			status = Stato.ERROR;
		}
		if (status != Stato.CONNECTED) return false;
		return true;
	}
	
	/**
	 * Richiede un blocco al <code>Peer</code> remoto
	 * @param blockIndex L'indice del blocco che si richiede al 
	 *        <code>Peer</code> remoto.
	 * @return <code>True</code> se non si sono verificati errori nel
	 *         mandare al <code>Peer</code> remoto il messaggio di richiesta
	 *         di un blocco, <code>False</code> altrimenti. */
	private synchronized boolean getBlock(int blockIndex){
		try {
			netOut.writeObject(Message.blockRequest(blockIndex));
			msg = (Message) netIn.readObject();
			switch(msg.code){
			case Message.BLOCK_ACK:
				Client.log.out("[TALKER] Ricevuto blocco "+msg.indice
					+" del file '"+filename+"' dal Peer "+peer.idPeer); 
				return swarm.filecache.put(msg.indice, msg.blocco);
			case Message.BLOCK_ERR:
				Client.log.err("[TALKER ERROR] Il Peer " +peer.idPeer+
					" non possiede il blocco '" +msg.indice+"' richiesto.");
				return false;
			default:
				Client.log.err("[TALKER ERROR] Messaggio '"+msg.code+
						"' non previsto. (reply)");
				status = Stato.ERROR;
				return false;
			}
		}
		catch (ClassNotFoundException e){
			Client.log.err("[TALKER ERROR] Oggetto messaggio" +
					" non riconosciuto.");
			Client.log.err(e);
			status = Stato.ERROR;
		}
		catch (EOFException ignore){status = Stato.REMOTELY_CLOSED;}
		catch (SocketException e){
			Client.log.err("[TALKER ERROR] Connessione remota chiusa con" +
					" il Peer "+peer.idPeer+" per il file '"+
					filename+"'. (requestBlock)");
			status = Stato.ERROR;
		}
		catch (IOException e){
			Client.log.err("[TALKER ERROR] Connession persa con il Peer "
					+peer.idPeer+". File '"+filename+"'. (getBlock) | " +
					e.getMessage());
			status = Stato.ERROR;
		}
		return false;
	}
	
	/**
	 * Richiede al <code>Peer</code> remoto il nuovo <code>BitSet</code>
	 * dei suoi blocchi posseduti.
	 * @return <code>True</code> se non si sono verificati errori nel
	 *         mandare al <code>Peer</code> remoto il messaggio di richiesta
	 *         del <code>BitSet</code> dei blocchi, <code>False</code>
	 *         altrimenti. */
	private synchronized boolean getBitSet(){
		try{
			netOut.reset();
			Client.log.out("[TALKER] Richiesto BitSet del Peer "+peer.idPeer
					+" per il file '"+filename+"' --->");
			netOut.writeObject(Message.getPeerBlocks());
			msg = (Message) netIn.readObject();
			switch(msg.code){
			case Message.BITSET_SEND:
				Client.log.out("[TALKER] Ricevuto BitSet dei blocchi dal"
						+" Peer "+peer.idPeer+" per il file '"+filename
						+"' <---");
				blocchi = msg.bs_blocchi;
				break;
			default:
				Client.log.err("[TALKER ERROR] Messaggio '"+msg.code+
						"' non previsto. (getPeerBlocks)");
				status = Stato.ERROR;
				break;			
			}
		}
		catch (ClassNotFoundException e){
			Client.log.err("[TALKER ERROR] Oggetto messaggio" +
					" non riconosciuto.");
			Client.log.err(e);
			status = Stato.ERROR;
		}
		catch (EOFException ignore){status = Stato.REMOTELY_CLOSED;}		
		catch (IOException e) {
			Client.log.err("[TALKER] Connessione chiusa con il Peer "
					+peer.idPeer+". File '"+filename+"' (getBitSet). | "
					+e.getMessage());
			status = Stato.ERROR;
		}
		if (status==Stato.CONNECTED) return true;
		else return false;
	}
	
	/**
	 * Chiude la connessione tra il <code>Peer</code> remoto e il
	 * <code>Peer</code> locale. */
	protected synchronized void closeConnection(){
		msg = null;
		try{
			if (status != Stato.REMOTELY_CLOSED){
				status = Stato.CLOSED;				
			}
			netIn.close();
			netOut.close();
			skClient.close();
		}
		catch (IOException e) {
			Client.log.err("[TALKER ERROR] Socket (closeConnection). | "
					+e.toString());
			status = Stato.ERROR;
		}
		/* Decremento il conteggio delle connessioni aperte */
		swarm.connection.decrementAndGet();
		inUse.getAndSet(false);
		synchronized (swarm.connection){swarm.connection.notify();}
	}
	
	/**
	 * Ritorna <code>True</code> se lo stato del canale di comunicazione
	 * e' chiuso.
	 * @return Ritorna <code>True</code> se lo stato del canale di
	 * comunicazione e' chiuso, <code>False</code> altrimenti. */
	protected boolean isClosed(){
		if (status == Stato.REMOTELY_CLOSED || status == Stato.CLOSED)
			return true;
		return false;
	}
	
	/**
	 * Ritorna <code>True</code> se il canale di comunicazione e'
	 * collegato al Peer remoto.
	 * @return Ritorna <code>True</code> se il canale di comunicazione
	 * collegato al Peer remoto, <code>False</code> altrimenti. */
	protected boolean isConnected(){
		return (status == Stato.CONNECTED);
	}
	
	/**
	 * Ritorna <code>True</code> se si e' verificato un errore di 
	 * comunicazione con il Peer remoto.
	 * @return Ritorna <code>True</code> se si e' verificato un errore di
	 * comunicazione con il Peer remoto, <code>False</code> altrimenti. */
	protected boolean isError(){
		return (status == Stato.ERROR);
	}
	
	/*___________________________________________________________________*/
	
	/**
	 * <h3>Thread per lo scariamento dei blocchi posseduti da questo Peer.
	 * </h3>
	 * Sara' il Peer a scaricare i blocchi in base alla lista di priorita'
	 * globale presente nello swarm e ai blocchi che lui possiede.<br>
	 * In questo modo e' possibile evitate attese attive sulla lista di
	 * priorita'.<br>
	 * Se un Peer non e' piu' in grado di dare un blocco utile, gli viene
	 * richiesto il nuovamente il suo BitSet dei blocchi aggiornato.<br>
	 * Se per due tentativi consecutivi il <code>Peer</code> non e' in grado
	 * di forinre nuovi blocchi utili, la socket viene chiusa e il 
	 * <code>Peer</code> rimosso dallo swarm locale. 
	 * Il risultato e' un BitSet dei blocchi che corrisponde solo ai
	 * nuovi blocchi che prima non erano posseduti dal Peer remoto.<br>
	 *  */
	public void run(){
		PriorityList pr = null;
		int tentativi = 0;
		
		while(!Client.shutdown && !shutdown){
			/* Controlla che la comunicazione sia effettivamente attiva */
			if (status != Stato.CONNECTED) break;
			pr = swarm.getRarestBlock(blocchi);
			if (swarm.filecache.testSeeder()) break;
			
			if (pr==null){
				if (tentativi>=2) break;
				/* Se il Peer e' un seeder e' inutile che gli chiedeo il
				 * suo bitset. */
				if (blocchi.cardinality() == swarm.nBlocchi) break;
				BitSet old_bitSet = (BitSet) blocchi.clone();
				/* Recupero il nuovo BitSet dei blocchi */
				if (!getBitSet()) break;
				old_bitSet.flip(0, swarm.filecache.getNumBlocchi());
				old_bitSet.and(blocchi);
				/* Aggiorno le strutture dati dello swarm */
				if (old_bitSet.cardinality()!=0)
					swarm.mergeBlocksList(old_bitSet);
				else
					tentativi++;
				
				/*Client.log.out("[TALKER DEBUG] Peer "+peer.idPeer
				 * 		+" AND = "+old_bitSet);*/
			}
			/* Recupero il blocco */
			else if (!getBlock(pr.indexBlock)){
				pr.download.set(false);
				shutdown = true;
				Client.log.err("[TALKER ERROR] Richiesta del blocco '"+
						pr.indexBlock+"' al Peer "+peer.idPeer+" fallita." );
			}
			/* Rimuovo il blocco dalla lista dei blocchi da scaricare */
			else{
				swarm.removeBlock(pr);
				tentativi = 0;
			}
		}
		/* Chiudo tutto */
		closeConnection();
		swarm.removePeer(this);
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "PeerTalk [blocchi=" + blocchi 
				+ "\nfilename = " + filename
				+ "\ninUse = " + inUse
				+ "\npeerId = " + peer.idPeer
				+ "\nshutdown = "	+ shutdown 
				+ "\nstatus = " + status 
				+ "\nswarm name = " + swarm.filecache.name()
				+ "]";
	}
}
