package bitcreek.server;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.rmi.AccessException;
import java.rmi.AlreadyBoundException;
import java.rmi.ConnectException;
import java.rmi.RemoteException;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import sun.misc.Signal;
import sun.misc.SignalHandler;

import bitcreek.Logger;

/**
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732<br>
 * 	     	Progetto LPR anno 2008/2009. */
public class Server implements SignalHandler {
	/**
	 * Millisecondi massimi di attesa prima dell'eliminazione di un Peer
	 * dallo swarm. Settare questo valore in stretta correlazione con il 
	 * tempo di invio di un Keepalive lato client.  */
	protected static int keepAliveTime = 0;
	/** Valore di default del KeepAlive espresso in millisecondi */
	protected final static int KEEPALIVE_DEF = 40000;
	/** Cartella che conterra' i file di log */
	private static final String LOG_DIR 		= "log";
	/** Dimensione massima di ogni file di log */
	private static final int MAX_SIZE			= 409600;	/* 400Kb */
	private static final String LOG_FILENAME	= "Server";
	/** Nome dello Stub */
	private static final String STUB_NAME 		= "BitCreek";
	/** Numero massimo di swarm per Tracker TCP di default */	
	public static final int SWARM_SIZE_DEF = 10;
	/** Numero massimo di swarm per Tracker TCP */	
	public static int trackerTCPswarmSize = 10;

	/** Oggetto per loggare su file e stampare a monitor ogni evento o
	 *  azione, e il loro relativo esito, intrapresa dal server. */
	public static Logger log =
		new Logger(LOG_DIR,LOG_FILENAME,MAX_SIZE,true);
	private static ServerRMIImpl server = null;
	private SignalHandler oldHandler;
	
    /* Gestione dei segnali */
    /** 
     * Installa il gestore dei segnali per il segnale <code>signalName</code>
     * passato alla funzione.
     * @param signalName Codice del segnale che si vuole gestire.
     * @return SignalHandler */
    public static SignalHandler install(String signalName) {
        Signal diagSignal = new Signal(signalName);
        Server instance = new Server();
        instance.oldHandler = Signal.handle(diagSignal, instance);
        return instance;
    }
	
    public void handle(Signal signal) {
    	log.err("[RMI] Segnale "+signal+" catturato.");
        try {
            signalAction(signal);
            if (oldHandler != SIG_DFL && oldHandler != SIG_IGN) {
                oldHandler.handle(signal);
            }
        }
        catch (Exception e) {
        	log.err("[RMI ERROR] handle Signal handler fallito."
        			+ e.getMessage());
            log.err(e);
        }
    }
    
    /**
     * Gestore dei segnali definito dall'utente per i segnali che si
     * vogliono gestire in modo specifico. 
     * @param signal Il segnale ricevuto. */    
    public void signalAction(Signal signal) {
		try {
			if (server!=null){
				server.shutdown();
				UnicastRemoteObject.unexportObject(server, false);
			}
		}
		catch (RemoteException e){log.err(e);}
        log.out("[RMI] Programma terminato.");
		log.flush();
    }
    
	/** Main
	 * @param args */
	public static void main(String[] args){
		int portaRMI = 0;

		/** Questi segnali sono gestiti correttamente
		 *  sia su piattaforma windows che linux. */
		Server.install("TERM");
		Server.install("INT");
		
		/* Recupero e setto il KeepAlive */		
		String keepAlive = System.getProperty("keepalive");
		if (keepAlive!=null){
			try{ keepAliveTime = Integer.parseInt(keepAlive); }
			catch(NumberFormatException e){
				keepAliveTime = KEEPALIVE_DEF;
			}
		}
		/* Non sono graditi valori negativi.. */		
		if (keepAliveTime<=0)
			keepAliveTime = KEEPALIVE_DEF;
		
		/* Recupero e setto la dimensione massima
		 * di ogni swarm per Tracker TCP */
		String swarmSize = System.getProperty("swarmssize");
		if (swarmSize!=null)
			try{ trackerTCPswarmSize = Integer.parseInt(swarmSize); }
			catch(NumberFormatException e){
				trackerTCPswarmSize = SWARM_SIZE_DEF;
			}
		if (trackerTCPswarmSize<=0)
			trackerTCPswarmSize = SWARM_SIZE_DEF;
		
		if (args.length==1){
			try{
				portaRMI = Integer.parseInt(args[0]);
				if (portaRMI < 1024 || portaRMI > 65535)
					throw new IllegalArgumentException("[RMI ERROR] valore"
							+ " non compreso nel range 1025-65535.");
			}
			catch(NumberFormatException e){
				portaRMI = java.rmi.registry.Registry.REGISTRY_PORT;
			}
		}
		else
			portaRMI = java.rmi.registry.Registry.REGISTRY_PORT;
		
		try {
			Registry registro = null;
			try{
				log.log("[RMI] Cerco di creare un rmiregistry.");
				registro =
					java.rmi.registry.LocateRegistry.createRegistry(portaRMI);
				log.out("[RMI] Registro RMI creato.");
			}
			catch(RemoteException e){
				log.log("[RMI ERROR] Registro gia' presente.");
				registro = java.rmi.registry.LocateRegistry.getRegistry();
				log.out("[RMI] Registro RMI recuperato.");
			}
			
			server = new ServerRMIImpl();
			
			try{
				log.log("[RMI] Eseguo la bind.");
				registro.bind(STUB_NAME, server);
				log.out("[RMI] Bind Ok.");
			}
			catch (RemoteException e1){
				log.err("[RMI ERROR] Bind fallita 'RemoteException'." +
						" Riprovo.");
				registro.rebind(STUB_NAME, server);
				log.out("[RMI] Bind Ok.");
			}
			catch (AlreadyBoundException e1){
				log.err("[RMI ERROR] Bind fallita 'AlreadyBoundException'."
						+ " Riprovo.");
				registro.rebind(STUB_NAME, server);
				log.out("[RMI] Bind Ok.");
			}
			log.out("[RMI] Server rmi avviato su "
					+InetAddress.getLocalHost()+":"+portaRMI
					+" [KeepAlive = "+keepAliveTime
					+", Swarm Size Tracker TCP = "
					+trackerTCPswarmSize+"]");
		}
		catch (IllegalArgumentException e){
			log.err(e.getMessage());
			if (server!=null) server.shutdown();
		}
		catch (ConnectException e){
			log.err("[RMI ERROR] Connessione al rmiregistry.");
			log.err(e);
			if (server!=null) server.shutdown();
		}
		catch (AccessException e){
			log.err("[RMI ERROR] Bind fallita 'AccessException'.");	
			log.err(e);
			if (server!=null) server.shutdown();
		}
		catch (RemoteException e){
			log.err("[RMI ERROR] "+e.getMessage());
			log.out("[RMI] Suggerimento. Si provi a cambiare la" +
					" porta del registro RMI.");
			if (server!=null) server.shutdown();
		}		
		catch (UnknownHostException e){
			log.err("[RMI ERROR]");
			log.err(e);
			if (server!=null) server.shutdown();			
		}
	}
}
