package bitcreek.server;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import bitcreek.Torrent;
import bitcreek.client.ClientInterface;

/**
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732<br>
 * 	     	Progetto LPR anno 2008/2009. */
public class ServerRMIImpl extends UnicastRemoteObject
implements ServerRMIInterface {
	private static final long serialVersionUID = 4423781303693435833L;
	/** Numero iniziale di thread */
	public static final int THREAD = 10;
	/** Numero massimo di thread */	
	public static final int MAX_THREAD = 100;
	/** Pool di thread del server */	
	private static ThreadPoolExecutor worker = null;
	/** Lista dei client che si sono registrati per ricevere le callback */
	private ConcurrentHashMap<String, ClientInterface> clients = null;
	/** HashTable dei descrittori dei file */
	private final ConcurrentHashMap<String, Torrent> torrentList
	 	= new ConcurrentHashMap<String, Torrent>();
	/** La lista dei tracker UDP */
	private Vector<TrackerUDP> trackers_udp = new Vector<TrackerUDP>();
	/** La lista dei tracker TCP */
	private ConcurrentHashMap<Integer, TrackerTCP> trackers_tcp =
		new ConcurrentHashMap<Integer, TrackerTCP>();
	
	
	/** Costruttore
	 * @throws RemoteException */
	public ServerRMIImpl() throws RemoteException {
		clients = new ConcurrentHashMap<String, ClientInterface>();
		worker = new ThreadPoolExecutor(THREAD,MAX_THREAD, 50000L,
				TimeUnit.MILLISECONDS, new LinkedBlockingQueue<Runnable>());
		/*worker.prestartAllCoreThreads();*/
	}
	
	/** Arresta dolcemente il server */
	public void shutdown(){
		worker.shutdown();
		
		synchronized(this){
			/* Arresto tutti i Tracker TCP per le richieste dello
			 * swarm dei descrittori di file da parte dei peer. */
			Enumeration<TrackerTCP> elems = trackers_tcp.elements();
			for (TrackerTCP tkTCP : Collections.list(elems))
				tkTCP.shutdown();
		
			/* Arresto tutti i Tracker UDP per il KeepAlive */
			for (TrackerUDP tkUDP : trackers_udp)
				tkUDP.shutdown();

			trackers_tcp.clear();
			trackers_udp.clear();
		}
		
		try {
			if (!worker.awaitTermination(15, TimeUnit.SECONDS)) {
				/* Cancel currently executing tasks */
				worker.shutdownNow();
				if (!worker.awaitTermination(15, TimeUnit.SECONDS))
					Server.log.err("[RMI ERROR] Il Pool di thread non e'" +
							" terminato correttamente.");
			}
		}
		catch (InterruptedException ie) {
			worker.shutdownNow();
			/* Preserva l'interrupt status */
			Thread.currentThread().interrupt();	
		}
	}
	
	/** Cerco una swarmList di un tracker ancora non piena 
	 * @return Una nuova swarmList */
	public synchronized SwarmList getSwarms(){
		Enumeration<TrackerTCP> elems = trackers_tcp.elements();
		for (TrackerTCP tkTCP : Collections.list(elems)){
			if (tkTCP.swarmList.size()<Server.trackerTCPswarmSize)
				return tkTCP.swarmList;
		}
		return null;
	}

	public synchronized Torrent publishDescriptor(InetSocketAddress idPeer, Torrent tor) {
		SwarmList swarms = null;
		boolean newTracker = false;
		
		try {
			/* Memorizzo sul server il nuovo descrittore */
			if (!torrentList.containsKey(tor.name))
				torrentList.put(tor.name, tor);
			else{
				Server.log.err("[RMI ERROR] Descrittore '"+tor.name
						+"' gia' esistente.");
				return null;
			}
			
			swarms = getSwarms();
			if (swarms == null){
				try{
					swarms = new SwarmList();
					newTracker = true;					
				}
				catch (Exception e){
					Server.log.err("[RMI ERROR] Swarm: "+e.getMessage());
					torrentList.remove(tor.name);
					Server.log.err(e);
					return null;
				}
			}

			/* Setto le porte del tracker sul torrent */
			tor.setTrackerPort(swarms.getTCPPort(), swarms.getUDPPort());
			
			/* Creo un nuovo swarm per il torrent e l'aggiungo */
			try { swarms.addSwarm(tor, idPeer); }
			catch (IOException e){
				torrentList.remove(tor.name);
				Server.log.err("[RMI ERROR] Swarm.");
				Server.log.err(e);
				return null;
			}
			
			/* Creo e attivo i nuovi tracker TCP e UDP */
			if (newTracker){
				TrackerUDP tracker_udp = new TrackerUDP(swarms);
				TrackerTCP tracker_tcp = new TrackerTCP(this, swarms);
				trackers_udp.add(tracker_udp);
				trackers_tcp.put(swarms.getTCPPort(), tracker_tcp);
				worker.execute(tracker_udp);
				worker.execute(tracker_tcp);
				Server.log.out("[RMI] Tracker UDP avviato sulla porta: "
					+swarms.getUDPPort()+".");
				Server.log.out("[RMI] Tracker TCP avviato sulla porta: "
					+swarms.getTCPPort()+".");				
			}
			Server.log.out("[RMI] Nuovo Descrittore '"+tor.name
					+"' pubblicato dal Peer "+ idPeer+".");
			return tor;	
		}
		catch (Exception e){
			Server.log.err("[RMI ERROR] "+e.getMessage());
			Server.log.err(e);
		}
		return null;
	}

	public synchronized Torrent getDescriptor(String filename, InetSocketAddress idPeer)
	throws RemoteException {
		Server.log.out("[RMI] Richiesta del descrittore per il file '"
				+filename+"' dal Peer "+idPeer+". <-");
		
		/* Controllo l'esistenza del descrittore di file richiesto */
		Torrent tor = torrentList.get(filename);
		if (tor==null){
			Server.log.err("[RMI ERROR] Descrittore del file '"+filename
					+"' richiesto dal Peer "+idPeer+" non trovato.");
			return null;
		}
		
		/* Controllo che esista un tracker tcp associato al descrittore */
		TrackerTCP tkTCP = trackers_tcp.get(tor.getTrackerTCPPort());
		if (tkTCP==null){
			Server.log.err("[RMI ERROR] Tracker TCP associato allo"
				+" swarm del descrittore '"+filename
				+"' richiesto dal Peer "+idPeer+" non trovato.");				
			return null;
		}

		/* Controllo che la sua lista di swarm non sia vuota */
		if (tkTCP.swarmList.size()==0){
			Server.log.err("[RMI ERROR] La SwarmList del Tracker TCP del "
				+"descrittore '"+filename+"' richiesto dal Peer "
				+idPeer+" e' vuota.");						
			return null;
		}

		/* Contollo l'esistenza effettiva dello swarm associato al
		 * descrittore */
		ServerSwarm swarm = tkTCP.swarmList.getSwarm(filename);
		if (swarm==null){
			Server.log.err("[RMI ERROR] Lo Swarm associato al descrittore '"
				+filename+"' richiesto dal Peer "+idPeer+" non esiste.");						
			return null;			
		}
		
		/* Rimuovo i peer vecchi */
		try { swarm.checkOldPeer();}
		catch (InterruptedException e){
			Server.log.err(e);
		}
		
		/* Controllo che lo swarm associato al descrittore richiesto
		 *  non sia vuoto */
		if (swarm.size()==0){
			Server.log.err("[RMI ERROR] Lo Swarm associato al descrittore '"
					+filename+"' richiesto dal Peer "+idPeer+" e' vuoto.");
			/* Rimuovo lo swarm dal tracker TCP */
			try {tkTCP.swarmList.removeSwarm(filename);}
			catch (InterruptedException e){
				Server.log.err(e);
			}
			/* Rimuovo il descrittore */
			torrentList.remove(filename);
			/* Rimuovo la callback associata al descrittore */
			clients.remove(filename);
			return null;
		}
		
		/* Effettua la CallBack verso il client che ha pubblicato
		 * il file se questo e' ancora registrato nella lista delle
		 * callback.*/
		ClientInterface pubblisher = clients.get(filename);
		if (pubblisher!=null){
			try{
				pubblisher.notifyMe(filename, idPeer);
				Server.log.out("[RMI] CallBack per il peer "
						+pubblisher.getId()+" per il file '"
						+filename+"' eseguita. ->");
			}
			catch(Exception e){
				Server.log.err("[RMI ERROR] Errore nel contattare il client"
					+pubblisher.getId()+" per effettuare la callback "
					+"per il file '"+filename+"'.");
			}
		}
		Server.log.out("[RMI] Descrittore del file '"+filename
				+"' inviato al Peer "+idPeer+". ->");
		return tor;
	}
	
	public synchronized void registerForCallback(String filename, ClientInterface client)
			throws RemoteException {
		if (!clients.containsKey(filename)){
			/* Memorizzo il peer nella lista delle callback */
			clients.put(filename, client);
			Server.log.out("[RMI] CallBack per il Peer "+
					client.getId()+" registrata.");
		}
		else 
			Server.log.err("[RMI ERROR] CallBack per il Peer = "+
					client.getId()+" per il descrittore '"+filename
					+"' non registrata. Il client si e' gia' registrato "
					+" sul server per questo file.");
	}
	
	public synchronized void unregisterForCallback(String filename, ClientInterface client)
			throws RemoteException {
		if (clients.remove(filename)!=null)
			Server.log.out("[RMI] CallBack cancellata per il Peer "
				+client.getId()+" per il descrittore '"
				+filename+"'.");
		else
			Server.log.err("[RMI ERROR] Callback non trovata per il Peer "+
				client.getId()+" per il descrittore '"+filename+"'.");
	}
	
	/**
	 * Rimuove un descrittore dal server e, se esite, la callback ad esso
	 * registrata.
	 * @param filename Il nome del descrittore che si vuole eliminare dal
	 *                 server. */
	protected synchronized void removeDescriptor(String filename){
		torrentList.remove(filename);
		clients.remove(filename);
	}
}
