package bitcreek.server;

import java.net.InetSocketAddress;
import java.rmi.Remote;
import java.rmi.RemoteException;

import bitcreek.Torrent;
import bitcreek.client.ClientInterface;

/**
 * Interfaccia del Server RMI che mette a disposizione del client le seguenti
 * funzioni:
 * <ul>
 *   <li><b><code>publishDesciptor</code></b> Permette ad un client di
 *       pubblicare un descrittore di file sul server.
 *   </li>
 *   <li><b><code>getDescriptor</code></b> Permette ad un client di
 *       recuperare un descritore di file pubblicato nel server.
 *   </li>
 *   <li><b><code>registerForCallback</code></b> Permette ad un client di
 *       registrarsi per ricevere le callback sul file cha ha pubblicato.
 *   </li>
 *   <li><b><code>unregisterForCallback</code></b> Operazione inversa della
 *       precedente, viene esguita dal client che ha pubblicato il<br> file
 *       quando ha deciso di lasciare lo swarm del file e non ha piu' la 
 *       necessita' di ricevere le callback. 
 *   </li>         
 * </ul>
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732<br>
 * 	     	Progetto LPR anno 2008/2009. */
public interface ServerRMIInterface extends Remote {

	/** 
	 * <h3>Pubblica un nuovo torrent sul server BitCreek:</h3>
	 * <ol>
	 *   <li>Controlla che non esista un descrittore con lo stesso nome
	 *       sul server.
	 *   </li>
	 *   <li>Cerca una swarmlist non satura a cui associare il nuovo
	 *       descrittore:
	 *     <ul>
	 *       <li>se non ne esiste una, creera' una nuova swarmlist a cui
	 *           associera' un nuovo tracker tcp e un nuovo tracker udp
	 *       </li>
	 *       <li>se esiste gia' una non satura, aggiungera' il descrittore
	 *           alla swarmlist trovata.
	 *       </li>
	 *     </ul>
	 *   </li>
	 *   <li>Aggiorna il descrittore col numero di porte del tracker tcp
	 *       e udp ad esso associato.
	 *   </li>
	 *   <li>Crea lo swarm per il descrittore e l'aggiunge alla swarmlist
	 *   </li>
	 *   <li>Aggiunge allo swarm appena creato il Peer che ha pubblicato il
	 *       file
	 *   </li>
	 * </ol>
	 * @param idPeer L'identita del client che sta pubblicando il
	 *         descrittore del file
	 * @param tor Il descrittore di file che si sta pubblicando sul server
	 * @return Torrent Ritorna il descrittore di file con i parametri
	 *         <code>portaTCP</code> e <code>portaUDP</code> settati dal
	 *         server, <code>null</code> se la pubblicazione sul server 
	 *         fallisce.
	 * @throws RemoteException */
	public Torrent publishDescriptor(InetSocketAddress idPeer, Torrent tor)
		throws RemoteException;
	
	/** Metodo per richiedere uno specifico Torrent al Server.
	 * La funzione esegue le seguenti operazioni nell'ordine:
	 * <ol>
	 *   <li>Controlla l'effettiva esistenza del descrittore nel server;
	 *   </li>
	 *   <li>Si accerta che esista un tracker tcp associato al descrittore;
	 *   </li>
	 *   <li>Controlla che la sua lista di swarm del tracker non sia
	 *       vuota;</li>
	 *   <li>Contolla l'esistenza effettiva dello swarm associato al
	 *       descrittore;</li>
	 *   <li>Rimuove i peer vecchi dallo swarm del descrittore</li>
	 *   <li>Controlla che lo swarm associato al descrittore richiesto non
	 *       sia vuoto, nel qual caso lo rimuove dalla lista degli swarm del
	 *       tracker e rimuove il descrittore di file dalla lista dei
	 *       descrittori pubblicati sul server;</li>
	 *   <li>Effettua la CallBack verso il client che ha pubblicato il file
	 *       se questo e' ancora registrato nella lista delle callback;</li>
	 *   <li>Restituisce il descrittore di file al peer che l'ha
	 *       richiesto;</li>
	 *</ol>
	 * In qualsiasi caso di errore, il metodo termina stampando a monitor
	 * un messaggio specifico e ritornando <code>null</code> al chiamante,
	 * altrimenti viene ritornato il descrittore di file e stampato a
	 * monitor un mesasggio di successo. 
	 * @param filename Il nome del torrent che si richiede al Server
	 * @param idPeer L'identita del Peer remoto che sta richiedendo il
	 *         descrittore di file
	 * @return <b><code>Torrent</code></b> Ritorna il torrent pubblicato
	 *         sul server
	 * @throws RemoteException */
	public Torrent getDescriptor (String filename, InetSocketAddress idPeer)
	throws RemoteException;
	
	/**
	 * Registra un client nella lista delle CallBack sul Server RMI per 
	 * un descittore di file specifico.
	 * @param filename Il nome del file a cui il client viene associato per
	 *        la CallBack
	 * @param client Il client che si registra per la CallBack
	 * @throws RemoteException Se si e' verificato un errore nel richiamare 
	 *        la procedura remota <code>cbClient.getClientId()</code>
	 *        del client. */
	public void registerForCallback(String filename, ClientInterface client)
		throws RemoteException; 
	
	/**
	 * Rimuove un client dalla lista delle CallBack del Server RMI per un 
	 * descrittore di file specifico.
	 * @param filename Il nome del file a cui il client e' stato associato
	 *        per la CallBack
	 * @param client Il client che si rimuove dalla lista delle
	 *        CallBack
	 * @throws RemoteException Se si e' verificato un errore nel richiamare 
	 *        la procedura remota <code>cbClient.getClientId()</code>
	 *        del client. */
	public void unregisterForCallback(String filename, ClientInterface client)
		throws RemoteException;
}
