package bitcreek.server;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.security.NoSuchAlgorithmException;
import java.util.Calendar;
import java.util.Collections;
import java.util.concurrent.ConcurrentHashMap;

import bitcreek.Peer;
import bitcreek.Torrent;

/**
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732<br>
 * 	     	Progetto LPR anno 2008/2009. */
public class ServerSwarm {
	/** HashTable dei Keepalive dei Peer dello Swarm */
	protected final ConcurrentHashMap<Integer, Calendar> keepAlive = 
		new ConcurrentHashMap<Integer, Calendar>();
	
	/** HashTable dei Peer dello Swarm */
	protected final ConcurrentHashMap<Integer, Peer> swarm =
		new ConcurrentHashMap<Integer, Peer>();
	
	/** Nome del file associato allo swarm */
	protected final String filename;
	
	/**
	 * Costruttore dello Swarm del Server.
	 * @param tor il descrittore del file a cui lo Swarm e' associato.
	 * @throws IOException 
	 * @throws NoSuchAlgorithmException */
	public ServerSwarm(Torrent tor)
	throws IOException, NoSuchAlgorithmException{
		filename = tor.name;
	}
	
	/**
	 * Aggiungo un nuovo peer nello swarm, senza specificare i blocchi che
	 * possiede (Di default non ne possiede nessuno).
	 * @param idPeer Oggetto 'InetSocketAddress' che identifica il peer.
	 * @return Il peer che e' stato aggiunto allo swarm.
	 * @throws NullPointerException */
	public synchronized Peer newPeer(InetSocketAddress idPeer)
	throws NullPointerException{
		Peer peer = null;
		if (idPeer!=null){	
			if (!swarm.containsKey(idPeer.hashCode())){
				peer = new Peer(idPeer);
				swarm.put(idPeer.hashCode(), peer);
				keepAlive.put(idPeer.hashCode(), Calendar.getInstance());
				Server.log.out("[Tracker TCP] Peer "+idPeer+
					" aggiunto nello swarm '"+filename+"'.");
			}
			else
				Server.log.out("[Tracker TCP] ERRORE Peer "+idPeer+
					" gia' presente nello swarm '"+filename+"'!");
		}
		else
			throw new NullPointerException("[Tracker TCP] ERRORE Peer" +
				" InetSocketAddress idPeer e' nullo!");
		return peer;
	}
	
	/**
	 * Controllo l'esistenza di un Peer nello Swarm.
	 * @param idPeer Oggetto 'InetSocketAddress' che identifica il peer.
	 * @return Il peer che e' stato aggiunto allo swarm.
	 * @throws NullPointerException */
	public synchronized boolean existPeer(InetSocketAddress idPeer)
	throws NullPointerException{
		if (idPeer!=null)
			return swarm.containsKey(idPeer.hashCode());
		else
			throw new NullPointerException("[Tracker TCP] ERRORE Peer" +
				" InetSocketAddress idPeer e' nullo!");
	}
	
	/**
	 * Ritorna l'hashtable dei Peer dello Swarm (TRACKER TCP).
	 * @return Ritorna l'hashtable dei Peer dello Swarm */
	public synchronized ConcurrentHashMap<Integer, Peer> getPeerList(){
		return swarm;
	}
	
	/**
	 * Aggiorno il keepAlive per uno specifico peer (x il TRACKER UDP)
	 * @param idPeer Oggetto 'InetSocketAddress' che identifica il peer.
	 * @throws InterruptedException
	 * @throws NullPointerException */
	public synchronized void refreshKeepAlive(InetSocketAddress idPeer)
	throws InterruptedException{
		Calendar keep_alive = keepAlive.get(idPeer.hashCode());
		if (keep_alive!=null){
			keepAlive.replace(idPeer.hashCode(), Calendar.getInstance());
		}
		else
			Server.log.err("[Tracker UDP] ERRORE Peer "+idPeer
					+" non trovato nello swarm!");
	}
	
	/**
	 * Rimuovo i peer troppo vecchi (x il TRACKER TCP)
	 * @throws InterruptedException */
	public synchronized void checkOldPeer()
	throws InterruptedException{
		for (int idPeer : Collections.list(keepAlive.keys())){
			Calendar now = Calendar.getInstance();
			now.add(Calendar.MILLISECOND, -Server.keepAliveTime);
			if (now.compareTo(keepAlive.get(idPeer))>0){
				keepAlive.remove(idPeer);
				Peer peer = swarm.remove(idPeer);
				if (peer!=null)
					Server.log.out("[Tracker TCP] Peer "+peer.idPeer
						+" rimosso dallo swarm del file '"+filename+"'.");
			}
		}		
	}
	
	/**
	 * Ritorna la dimensione dello swarm
	 * @return La dimensione dello Swarm */
	public synchronized int size(){
		return swarm.size();
	}
}
