package bitcreek.server;

import java.io.FileInputStream;
import java.io.IOException;
import java.util.concurrent.ConcurrentHashMap;
import java.net.BindException;
import java.net.DatagramSocket;
import java.net.InetSocketAddress;
import java.security.KeyStore;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLServerSocket;
import javax.net.ssl.SSLServerSocketFactory;
import javax.net.ssl.SSLSocket;

import bitcreek.Torrent;

/**
 * @author  Giulio Tranchida (gtranchi) no. matricola 241732<br>
 * 	     	Progetto LPR anno 2008/2009. */
public class SwarmList {
	/** Lista degli Swarms per il Server RMI e i Trackers */
	private ConcurrentHashMap<String, ServerSwarm> swarmList = null;
	/** Socket di ricezione del Tracker TCP */
	private SSLServerSocket sslServerSK = null;
	SSLServerSocketFactory sslServerSKFactory = null;
	/** Socket di ricezione del Tracker UDP */
	private DatagramSocket skUDP = null;
	
	/** 
    * Costruttore della classe.
    * Crea le socket di comunicazione per il tracker TCP (ssl) e
    * il tracker UDP e inizializza l'hashatable swarmlist.
	 * @throws Exception */
	public SwarmList() throws Exception{
		swarmList = new ConcurrentHashMap<String, ServerSwarm>();
		sslServerSKFactory = getServerSocketFactory();
		/* Cerco una socket per il Tracker TCP */
		for (int i = 1025; i<65535; i++){
			try {
				sslServerSK = (SSLServerSocket)
							sslServerSKFactory.createServerSocket(i);
				break;
			}
			catch (BindException e){/* Porta gia' occupata */} 
		}
		/* Cerco una socket per il Tracker UDP */
		for (int i = 1025; i<65535; i++){
			try {
				skUDP = new DatagramSocket(i);
				break;
			}
			catch (BindException e){/* Porta gia' occupata */} 
		}
	}
	
	/** 
	 * Restituisce un SSLServerSocketFactory per la socket SSL del Tracker TCP 
	 * @see SSLSocket
	 * @see SSLServerSocketFactory
	 * @return SSLServerSocketFactory 
	 * @throws Exception */
	private static SSLServerSocketFactory getServerSocketFactory()
	throws Exception{
		SSLContext ctx = SSLContext.getInstance("SSL");
		KeyManagerFactory kmf =
			KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
		KeyStore ks = KeyStore.getInstance("JKS");

		char[] passphrase = "password".toCharArray();
		ks.load(new FileInputStream("bitcreek_server.jks"), passphrase);
		kmf.init(ks, passphrase);
		ctx.init(kmf.getKeyManagers(), null, new SecureRandom()); 
		return ctx.getServerSocketFactory();
	}
	
	/** <h3>Aggiunge uno swarm alla lista degli swarm e lo inizializza.</h3>
	 * Quando viene pubblicato un nuovo descrittore, il server RMI
	 * a) creera' uno nuovo swarm per il nuovo file e aggiungera',
	 * b) come primo peer dello swarm, il client stesso che ha pubblicato
	 *    il file.
	 * @param tor Il torrent dello swarm 
	 * @param idPeer L'InetSocketAddress univoco del client
	 * @return Lo swarm appena creato
	 * @throws InterruptedException 
	 * @throws IOException 
	 * @throws NoSuchAlgorithmException */
	public synchronized ServerSwarm addSwarm(Torrent tor, InetSocketAddress idPeer)
	throws InterruptedException, IOException, NoSuchAlgorithmException{
		ServerSwarm swarm = new ServerSwarm(tor);
		swarm.newPeer(idPeer);
		swarmList.put(tor.name, swarm);
		return swarm;
	}
	
	/** <h3>Aggiorna il KeepAlive del Peer, per un paricolare swarm.</h3>
	 * Questa funzione viene richiamata solo dal Tracker UDP, quando questo
	 * riceve un KeepAlive da un peer appartenente ad uno dei suoi swarm.
	 * @param name Il nome dello swarm 
	 * @param idPeer L'InetSocketAddress univoco del client
	 * @throws NullPointerException Se <code>name</code> e' <code>null<code>.
	 * @throws InterruptedException */
	public synchronized void keepAlive(String name, InetSocketAddress idPeer)
	throws InterruptedException, NullPointerException{
		if (name==null)
			throw new NullPointerException("[KeepAlive] ERRORE Peer "+
				"String name e' nulla!");
		if (swarmList.containsKey(name))
			swarmList.get(name).refreshKeepAlive(idPeer);
		else
			Server.log.err("[KeepAlive] ERRORE Swarm '"+name+
					"' non presente!");
	}
	
	/** <h3>Rimuove uno swarm.</h3>
	 * Questa funzione viene richiamata solo dal Tracker TCP.
	 * @param name Nome dello Swarm da rimuovere
	 * @throws InterruptedException */
	public synchronized void removeSwarm(String name)
	throws InterruptedException{
		ServerSwarm swarm = swarmList.remove(name);
		if (swarm == null)
			Server.log.err("[Tracker TCP] ERRORE Swarm '"+name
					+"' non presente!");
		else{
			swarm.keepAlive.clear();
			swarm.swarm.clear();
			Server.log.out("[Tracker TCP] Swarm '"+name+"' rimosso.");
		}
	}
	
	/** Ritorna uno swarm appartente alla <code>SwarmList</code>.
	 * @param nome Nome dello Swarm
	 * @return Ritorna lo swarm appartente alla swarmlist con chiave 'nome' */
	public ServerSwarm getSwarm(String nome){
		return swarmList.get(nome);
	}
	
	/** Ritorna la dimesione della <code>SwarmList</code>.
	 * @return la dimesione dell'hashtable */
	public int size(){
		return swarmList.size();
	}

	/**
	 * @return the sslServerSK */
	public SSLServerSocket getSslServerSK() {
		return sslServerSK;
	}

	/**
	 * @return the skUDP */
	public DatagramSocket getSkUDP() {
		return skUDP;
	}
	
	/** @return Ritorna la porta della socket TCP */
	public int getTCPPort(){
		return getSslServerSK().getLocalPort();
	}
	/** @return Ritorna la porta della socket UDP */
	public int getUDPPort(){
		return getSkUDP().getLocalPort(); 
	}
}
